import inspect
from typing import Optional, Type

from arcor2.clients import persistent_storage as storage
from arcor2.data.object_type import Models, ObjectType
from arcor2.exceptions import Arcor2Exception
from arcor2.helpers import ImportClsException, import_cls
from arcor2.object_types.abstract import Generic, GenericWithPose
from arcor2.object_types.utils import check_object_type


class UploadException(Arcor2Exception):
    pass


def upload_def(type_def: Type[Generic], model: Optional[Models] = None) -> None:

    if not issubclass(type_def, GenericWithPose) and model:
        raise UploadException("Object without pose can't have collision model.")

    path = inspect.getfile(type_def)

    try:
        check_object_type(type_def)
    except Arcor2Exception as e:
        print(e)
        raise UploadException(f"There is something wrong with source code of '{type_def.__name__}'.")

    with open(path, "r") as source_file:

        source = source_file.read()

    obj_type = ObjectType(id=type_def.__name__, source=source, desc=type_def.description())

    if model:
        obj_type.model = model.metamodel()
        storage.put_model(model)

    print(f"Storing '{obj_type.id}'...")
    storage.update_object_type(obj_type)


def upload(module_cls: str, model: Optional[Models] = None) -> None:
    try:
        _, cls = import_cls(module_cls)
    except ImportClsException as e:
        raise UploadException(e)

    if not issubclass(cls, Generic):
        raise UploadException("Object should be derived from Generic.")

    upload_def(cls, model)



#!/usr/bin/env python3

import uuid
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum, unique
from json import JSONEncoder
from typing import Any, ClassVar, Iterator, List, NamedTuple, Optional, Set, cast

import quaternion  # type: ignore
from dataclasses_jsonschema import JsonSchemaMixin

from arcor2.exceptions import Arcor2Exception


def uid() -> str:
    return uuid.uuid4().hex


@unique
class StrEnum(str, Enum):
    @classmethod
    def set(cls) -> Set[str]:
        return set(map(lambda c: c.value, cls))  # type: ignore


@unique
class IntEnum(int, Enum):
    @classmethod
    def set(cls) -> Set[int]:
        return set(map(lambda c: c.value, cls))  # type: ignore


class FlowTypes(StrEnum):

    DEFAULT: str = "default"


class LinkToActionOutput(NamedTuple):
    action_id: str
    flow_name: FlowTypes
    output_index: int


def parse_link(val: str) -> LinkToActionOutput:

    action_id, flow_name, output_idx_str = val.split("/")

    try:
        return LinkToActionOutput(action_id, FlowTypes(flow_name), int(output_idx_str))
    except ValueError as e:
        raise Arcor2Exception("Invalid link value.") from e


class DataClassEncoder(JSONEncoder):
    def default(self, o: Any) -> Any:

        if isinstance(o, JsonSchemaMixin):
            return o.to_dict()

        return super().default(o)


@dataclass
class IterableIndexable(JsonSchemaMixin):
    def __getitem__(self, item: int) -> float:

        attr = getattr(self, tuple(self.__dict__.keys())[item])
        assert isinstance(attr, float)
        return attr

    def __iter__(self) -> Iterator[float]:

        yield from self.__dict__.values()


@dataclass
class Position(IterableIndexable):

    x: float = 0.0
    y: float = 0.0
    z: float = 0.0

    def rotated(self, rot: "Orientation", inverse: bool = False) -> "Position":

        q = rot.as_quaternion()

        if inverse:
            q = q.inverse()

        rotated_vector = quaternion.rotate_vectors([q], [list(self)])[0][0]
        return Position(rotated_vector[0], rotated_vector[1], rotated_vector[2])


@dataclass
class Orientation(IterableIndexable):

    x: float = 0.0
    y: float = 0.0
    z: float = 0.0
    w: float = 1.0

    @staticmethod
    def _normalized(q: quaternion.quaternion) -> quaternion.quaternion:

        nq = q.normalized()

        if nq.isnan():
            raise Arcor2Exception("Invalid quaternion.")

        return nq

    def as_quaternion(self) -> quaternion.quaternion:
        return self._normalized(quaternion.quaternion(self.w, self.x, self.y, self.z))

    def set_from_quaternion(self, q: quaternion.quaternion) -> None:

        nq = self._normalized(q)

        self.x = nq.x
        self.y = nq.y
        self.z = nq.z
        self.w = nq.w

    def __eq__(self, other: object) -> bool:

        if not isinstance(other, Orientation):
            return False

        return cast(bool, quaternion.isclose(self.as_quaternion(), other.as_quaternion(), rtol=1.0e-8)[0])

    def __post_init__(self):

        nq = self.as_quaternion()  # in order to get normalized quaternion

        self.x = nq.x
        self.y = nq.y
        self.z = nq.z
        self.w = nq.w


@dataclass
class NamedOrientation(JsonSchemaMixin):

    id: str
    name: str
    orientation: Orientation


@dataclass
class Pose(JsonSchemaMixin):

    position: Position = field(default_factory=Position)
    orientation: Orientation = field(default_factory=Orientation)


@dataclass
class ActionMetadata(JsonSchemaMixin):

    blocking: bool = False
    composite: bool = False
    blackbox: bool = False
    cancellable: bool = field(init=False, default=False)


@dataclass
class Joint(JsonSchemaMixin):

    name: str
    value: float


@dataclass
class ProjectRobotJoints(JsonSchemaMixin):

    id: str
    name: str
    robot_id: str
    joints: List[Joint]
    is_valid: bool = False


@dataclass
class Parameter(JsonSchemaMixin):

    name: str
    type: str
    value: str


@dataclass
class SceneObject(JsonSchemaMixin):

    id: str
    name: str
    type: str
    pose: Optional[Pose] = None
    parameters: List[Parameter] = field(default_factory=list)
    children: List["SceneObject"] = field(default_factory=list)


@dataclass
class BareScene(JsonSchemaMixin):

    id: str
    name: str
    desc: str = field(default_factory=str)
    modified: Optional[datetime] = None
    int_modified: Optional[datetime] = None


@dataclass
class Scene(BareScene):

    objects: List[SceneObject] = field(default_factory=list)

    @staticmethod
    def from_bare(bare: BareScene) -> "Scene":
        return Scene(bare.id, bare.name, bare.desc, bare.modified, bare.int_modified)


@dataclass
class IdValue(JsonSchemaMixin):

    id: str
    value: str


class ActionParameterException(Arcor2Exception):
    pass


@dataclass
class ActionParameter(Parameter):
    class TypeEnum(StrEnum):

        CONSTANT: str = "constant"
        LINK: str = "link"

    def parse_link(self) -> LinkToActionOutput:
        assert self.type == self.TypeEnum.LINK
        return parse_link(self.value)

    def is_value(self) -> bool:
        return self.type not in self.TypeEnum.set()


@dataclass
class Flow(JsonSchemaMixin):

    type: FlowTypes = FlowTypes.DEFAULT
    outputs: List[str] = field(default_factory=list)  # can't be set as it is unordered

    def __post_init__(self) -> None:

        if len(self.outputs) > len(set(self.outputs)):
            raise Arcor2Exception("Outputs have to be unique.")


@dataclass
class BareAction(JsonSchemaMixin):

    id: str
    name: str
    type: str


@dataclass
class Action(BareAction):
    class ParsedType(NamedTuple):

        obj_id: str
        action_type: str

    parameters: List[ActionParameter] = field(default_factory=list)
    flows: List[Flow] = field(default_factory=list)

    def parse_type(self) -> ParsedType:

        try:
            obj_id_str, action = self.type.split("/")
        except ValueError:
            raise Arcor2Exception(f"Action: {self.id} has invalid type: {self.type}.")
        return self.ParsedType(obj_id_str, action)

    def parameter(self, parameter_id: str) -> ActionParameter:

        for param in self.parameters:
            if parameter_id == param.name:
                return param

        raise Arcor2Exception("Param not found")

    @property
    def bare(self) -> BareAction:
        return BareAction(self.id, self.name, self.type)

    def flow(self, flow_type: FlowTypes = FlowTypes.DEFAULT) -> Flow:

        for flow in self.flows:
            if flow.type == flow_type:
                return flow
        raise Arcor2Exception(f"Flow '{flow_type.value}' not found.")


@dataclass
class BareActionPoint(JsonSchemaMixin):

    id: str
    name: str
    position: Position
    parent: Optional[str] = None


@dataclass
class ActionPoint(BareActionPoint):

    orientations: List[NamedOrientation] = field(default_factory=list)
    robot_joints: List[ProjectRobotJoints] = field(default_factory=list)
    actions: List[Action] = field(default_factory=list)

    @staticmethod
    def from_bare(bare: BareActionPoint) -> "ActionPoint":
        return ActionPoint(bare.id, bare.name, bare.position, bare.parent)


@dataclass
class ProjectLogicIf(JsonSchemaMixin):

    what: str
    value: str

    def parse_what(self) -> LinkToActionOutput:
        return parse_link(self.what)


@dataclass
class LogicItem(JsonSchemaMixin):
    class ParsedStart(NamedTuple):

        start_action_id: str
        start_flow: str

    START: ClassVar[str] = "START"
    END: ClassVar[str] = "END"

    id: str
    start: str
    end: str
    condition: Optional[ProjectLogicIf] = None

    def parse_start(self) -> ParsedStart:

        try:
            start_action_id, start_flow = self.start.split("/")
        except ValueError:
            return self.ParsedStart(self.start, FlowTypes.DEFAULT)

        return self.ParsedStart(start_action_id, start_flow)


@dataclass
class ProjectConstant(JsonSchemaMixin):

    id: str
    name: str
    type: str
    value: str


@dataclass
class FunctionReturns(JsonSchemaMixin):

    type: str
    link: str


@dataclass
class ProjectFunction(JsonSchemaMixin):

    id: str
    name: str
    actions: List[Action] = field(default_factory=list)
    logic: List[LogicItem] = field(default_factory=list)
    parameters: List[ActionParameter] = field(default_factory=list)
    returns: List[FunctionReturns] = field(default_factory=list)

    def action_ids(self) -> Set[str]:
        return {act.id for act in self.actions}

    def action(self, action_id: str) -> Action:

        for ac in self.actions:
            if ac.id == action_id:
                return ac
        else:
            raise Arcor2Exception("Action not found")


@dataclass
class SceneObjectOverride(JsonSchemaMixin):

    id: str
    parameters: List[Parameter]


@dataclass
class BareProject(JsonSchemaMixin):

    id: str
    name: str
    scene_id: str
    desc: str = field(default_factory=str)
    has_logic: bool = True
    modified: Optional[datetime] = None
    int_modified: Optional[datetime] = None


@dataclass
class Project(BareProject):

    action_points: List[ActionPoint] = field(default_factory=list)
    constants: List[ProjectConstant] = field(default_factory=list)
    functions: List[ProjectFunction] = field(default_factory=list)
    logic: List[LogicItem] = field(default_factory=list)
    object_overrides: List[SceneObjectOverride] = field(default_factory=list)

    @staticmethod
    def from_bare(bare: BareProject) -> "Project":
        return Project(bare.id, bare.name, bare.scene_id, bare.desc, bare.has_logic, bare.modified, bare.int_modified)


@dataclass
class ProjectSources(JsonSchemaMixin):

    id: str  # project_id
    script: str


@dataclass
class IdDesc(JsonSchemaMixin):
    id: str
    name: str
    desc: Optional[str]


@dataclass
class IdDescList(JsonSchemaMixin):

    items: List[IdDesc] = field(default_factory=list)


@dataclass
class BroadcastInfo(JsonSchemaMixin):

    host: str
    port: int



import time
from typing import Optional

from arcor2.data.common import ActionMetadata
from arcor2.object_types.abstract import Generic, Settings


class TimeActions(Generic):
    """Time-related actions."""

    _ABSTRACT = False

    def __init__(self, obj_id: str, name: str, settings: Optional[Settings] = None) -> None:
        super(TimeActions, self).__init__(obj_id, name, settings)
        self._last_time: Optional[float] = None

    def sleep(self, seconds: float = 1.0) -> None:

        assert 0.0 <= seconds <= 10.0e6

        time.sleep(seconds)

    def rate(self, period: float = 1.0) -> None:
        """Can be used to maintain desired rate of the loop. Should be the
        first action.

        :param period: Desired period in seconds.
        :return:
        """

        assert 0.0 <= period <= 10.0e6

        now = time.monotonic()

        if self._last_time is not None:

            dif = self._last_time + period - now

            if dif > 0:
                time.sleep(dif)

        self._last_time = now

    def time_ns(self) -> int:
        """Returns nanoseconds since Unix Epoch.

        :return:
        """
        return time.time_ns()

    sleep.__action__ = ActionMetadata(blocking=True)  # type: ignore
    rate.__action__ = ActionMetadata(blocking=True)  # type: ignore
    time_ns.__action__ = ActionMetadata(blocking=True)  # type: ignore



import abc
import copy
import inspect
from dataclasses import dataclass
from typing import List, Optional, Set

from dataclasses_jsonschema import JsonSchemaMixin

from arcor2 import CancelDict, DynamicParamDict
from arcor2.clients import scene_service
from arcor2.data.common import Joint, Pose, SceneObject
from arcor2.data.object_type import Models
from arcor2.data.robot import RobotType
from arcor2.docstring import parse_docstring
from arcor2.exceptions import Arcor2Exception
from arcor2.helpers import NonBlockingLock


class GenericException(Arcor2Exception):
    pass


@dataclass
class Settings(JsonSchemaMixin):
    """Default (empty) settings for ARCOR2 objects."""

    pass


class Generic(metaclass=abc.ABCMeta):
    """Generic object."""

    DYNAMIC_PARAMS: DynamicParamDict = {}
    CANCEL_MAPPING: CancelDict = {}
    _ABSTRACT = True

    def __init__(self, obj_id: str, name: str, settings: Optional[Settings] = None) -> None:

        self.id = obj_id
        self.name = name

        if settings is None:
            settings = Settings()
        self._settings = settings

    @classmethod
    def abstract(cls) -> bool:
        return inspect.isabstract(cls) or cls._ABSTRACT

    @property
    def settings(self) -> Settings:
        return self._settings

    @classmethod
    def description(cls) -> str:
        if not cls.__doc__:
            return "No description available."
        return parse_docstring(cls.__doc__)["short_description"]

    def scene_object(self) -> SceneObject:
        return SceneObject(self.id, self.name, self.__class__.__name__)

    def __repr__(self) -> str:
        return str(self.__dict__)

    def cleanup(self) -> None:
        """This method is called when a scene is closed or when script ends.

        :return:
        """
        pass


class GenericWithPose(Generic):
    def __init__(
        self,
        obj_id: str,
        name: str,
        pose: Pose,
        collision_model: Optional[Models] = None,
        settings: Optional[Settings] = None,
    ) -> None:

        super(GenericWithPose, self).__init__(obj_id, name, settings)

        self._pose = pose
        self.collision_model = copy.deepcopy(collision_model)
        if self.collision_model:
            # originally, each model has id == object type (e.g. BigBox) but here we need to set it to something unique
            self.collision_model.id = self.id
            scene_service.upsert_collision(self.collision_model, pose)

    def scene_object(self) -> SceneObject:
        return SceneObject(self.id, self.name, self.__class__.__name__, self._pose)

    @property
    def pose(self) -> Pose:
        return self._pose

    @pose.setter
    def pose(self, pose: Pose) -> None:
        self._pose = pose
        if self.collision_model:
            scene_service.upsert_collision(self.collision_model, pose)

    def cleanup(self) -> None:

        super(GenericWithPose, self).cleanup()

        if self.collision_model:
            scene_service.delete_collision_id(self.collision_model.id)


class RobotException(Arcor2Exception):
    pass


class Robot(GenericWithPose, metaclass=abc.ABCMeta):
    """Abstract class representing robot and its basic capabilities (motion)"""

    def __init__(self, obj_id: str, name: str, pose: Pose, settings: Optional[Settings] = None) -> None:
        super(Robot, self).__init__(obj_id, name, pose, None, settings)
        self._move_lock = NonBlockingLock()

    robot_type = RobotType.ARTICULATED
    urdf_package_name: Optional[str] = None

    @property
    def move_in_progress(self) -> bool:
        return self._move_lock.locked()

    def check_if_ready_to_move(self) -> None:
        """The method should raise exception when the robot is not ready to
        move for some reason.

        :return:
        """

        if self.move_in_progress:
            raise RobotException("Already moving.")

        return None

    @abc.abstractmethod
    def get_end_effectors_ids(self) -> Set[str]:
        pass

    @abc.abstractmethod
    def get_end_effector_pose(self, end_effector: str) -> Pose:
        pass

    @abc.abstractmethod
    def robot_joints(self) -> List[Joint]:
        pass

    @abc.abstractmethod
    def grippers(self) -> Set[str]:
        return set()

    @abc.abstractmethod
    def suctions(self) -> Set[str]:
        return set()

    def move_to_pose(self, end_effector_id: str, target_pose: Pose, speed: float) -> None:
        """Move given robot's end effector to the selected pose.

        :param end_effector_id:
        :param target_pose:
        :param speed:
        :return:
        """

        assert 0.0 <= speed <= 1.0
        raise NotImplementedError("Robot does not support moving to pose.")

    def move_to_joints(self, target_joints: List[Joint], speed: float) -> None:
        """Sets target joint values.

        :param target_joints:
        :param speed:
        :return:
        """

        assert 0.0 <= speed <= 1.0
        raise NotImplementedError("Robot does not support moving to joints.")

    def stop(self) -> None:
        raise NotImplementedError("The robot can't be stopped.")


__all__ = [
    Generic.__name__,
    GenericWithPose.__name__,
    Robot.__name__,
    GenericException.__name__,
    RobotException.__name__,
]



from typing import Any

from arcor2.cached import CachedProject as CProject
from arcor2.cached import CachedScene as CScene
from arcor2.data.common import ProjectRobotJoints
from arcor2.parameter_plugins.base import ParameterPlugin, ParameterPluginException, TypesDict


class JointsPlugin(ParameterPlugin):
    @classmethod
    def type(cls) -> Any:
        return ProjectRobotJoints

    @classmethod
    def type_name(cls) -> str:
        return "joints"

    @classmethod
    def value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> ProjectRobotJoints:

        ap, action = project.action_point_and_action(action_id)
        param = action.parameter(parameter_id)
        joints_id = cls.param_value(param)

        robot_id, action_method_name = action.parse_type()

        joints = project.joints(joints_id)

        if joints.robot_id != robot_id:
            raise ParameterPluginException("Joints are for different robot.")

        return joints

    @classmethod
    def value_to_json(cls, value: ProjectRobotJoints) -> str:
        return value.to_json()

    @classmethod
    def uses_robot_joints(cls, project: CProject, action_id: str, parameter_id: str, robot_joints_id: str) -> bool:

        param = project.action(action_id).parameter(parameter_id)
        value_id = cls.param_value(param)

        return value_id == robot_joints_id



import logging

import aiologger  # type: ignore

LOG_FORMAT = "%(asctime)s - %(levelname)-8s: %(message)s", "%Y-%m-%d %H:%M:%S"


def aiologger_formatter() -> aiologger.formatters.base.Formatter:

    return aiologger.formatters.base.Formatter(*LOG_FORMAT)


def logger_formatter() -> logging.Formatter:

    return logging.Formatter(*LOG_FORMAT)


def get_aiologger(name: str, level: aiologger.levels.LogLevel = aiologger.levels.LogLevel.INFO) -> aiologger.Logger:

    return aiologger.Logger.with_default_handlers(name=name, formatter=aiologger_formatter(), level=level)


def get_logger(name: str, level: int = logging.INFO) -> logging.Logger:

    logger = logging.getLogger(name)

    ch = logging.StreamHandler()
    ch.setFormatter(logger_formatter())
    logger.setLevel(level)
    logger.addHandler(ch)

    return logger



from arcor2.cached import CachedProject as CProject
from arcor2.cached import CachedScene as CScene
from arcor2.data.common import BareActionPoint, Orientation, Pose, Position
from arcor2.exceptions import Arcor2Exception


def make_position_rel(parent: Position, child: Position) -> Position:

    p = Position()

    p.x = child.x - parent.x
    p.y = child.y - parent.y
    p.z = child.z - parent.z
    return p


def make_orientation_rel(parent: Orientation, child: Orientation) -> Orientation:

    p = Orientation()
    p.set_from_quaternion(child.as_quaternion() / parent.as_quaternion())
    return p


def make_pose_rel(parent: Pose, child: Pose) -> Pose:
    """
    :param parent: e.g. scene object
    :param child:  e.g. action point
    :return: relative pose
    """

    p = Pose()
    p.position = make_position_rel(parent.position, child.position).rotated(parent.orientation, True)
    p.orientation = make_orientation_rel(parent.orientation, child.orientation)
    return p


def make_position_abs(parent: Position, child: Position) -> Position:

    p = Position()
    p.x = child.x + parent.x
    p.y = child.y + parent.y
    p.z = child.z + parent.z
    return p


def make_orientation_abs(parent: Orientation, child: Orientation) -> Orientation:

    p = Orientation()
    p.set_from_quaternion(child.as_quaternion() * parent.as_quaternion().conjugate().inverse())
    return p


def make_pose_abs(parent: Pose, child: Pose) -> Pose:
    """
    :param parent: e.g. scene object
    :param child:  e.g. action point
    :return: absolute pose
    """

    p = Pose()
    p.position = child.position.rotated(parent.orientation)
    p.position = make_position_abs(parent.position, p.position)
    p.orientation = make_orientation_abs(parent.orientation, child.orientation)
    return p


def make_relative_ap_global(scene: CScene, project: CProject, ap: BareActionPoint) -> None:
    """Transforms (in place) relative AP into a global one.

    :param scene:
    :param project:
    :param ap:
    :return:
    """

    if not ap.parent:
        return

    if ap.parent in scene.object_ids:
        parent_obj = scene.object(ap.parent)
        if not parent_obj.pose:
            raise Arcor2Exception("Parent object does not have pose!")
        old_parent_pose = parent_obj.pose
    elif ap.parent in project.action_points_ids:
        old_parent_pose = Pose(project.bare_action_point(ap.parent).position, Orientation())
    else:
        raise Arcor2Exception("AP has unknown parent_id.")

    ap.position = make_pose_abs(old_parent_pose, Pose(ap.position, Orientation())).position
    for ori in project.ap_orientations(ap.id):
        ori.orientation = make_orientation_abs(old_parent_pose.orientation, ori.orientation)

    if ap.parent in project.action_points_ids:
        parent_ap = project.bare_action_point(ap.parent)
        if parent_ap.parent:
            ap.parent = parent_ap.parent
            make_relative_ap_global(scene, project, ap)

    ap.parent = None


def make_global_ap_relative(scene: CScene, project: CProject, ap: BareActionPoint, parent_id: str) -> None:
    """Transforms (in place) global AP into a relative one with given parent
    (can be object or another AP).

    :param scene:
    :param project:
    :param ap:
    :param parent_id:
    :return:
    """

    assert project.scene_id == scene.id

    if parent_id in scene.object_ids:
        parent_obj = scene.object(parent_id)
        if not parent_obj.pose:
            raise Arcor2Exception("Parent object does not have pose!")
        new_parent_pose = parent_obj.pose
    elif parent_id in project.action_points_ids:

        parent_ap = project.bare_action_point(parent_id)

        if parent_ap.parent:
            make_global_ap_relative(scene, project, ap, parent_ap.parent)

        new_parent_pose = Pose(parent_ap.position, Orientation())

    else:
        raise Arcor2Exception("Unknown parent_id.")

    ap.position = make_pose_rel(new_parent_pose, Pose(ap.position, Orientation())).position
    for ori in project.ap_orientations(ap.id):
        ori.orientation = make_orientation_rel(new_parent_pose.orientation, ori.orientation)

    ap.parent = parent_id


def make_pose_rel_to_parent(scene: CScene, project: CProject, pose: Pose, parent_id: str) -> Pose:
    """Transforms global Pose into Pose that is relative to a given parent (can
    be object or AP).

    :param scene:
    :param project:
    :param pose:
    :param parent_id:
    :return:
    """

    if parent_id in scene.object_ids:
        parent_obj = scene.object(parent_id)
        if not parent_obj.pose:
            raise Arcor2Exception("Parent object does not have pose!")
        parent_pose = parent_obj.pose
    elif parent_id in project.action_points_ids:

        parent_ap = project.bare_action_point(parent_id)

        if parent_ap.parent:
            pose = make_pose_rel_to_parent(scene, project, pose, parent_ap.parent)

        parent_pose = Pose(parent_ap.position, Orientation())

    else:
        raise Arcor2Exception("Unknown parent_id.")

    return make_pose_rel(parent_pose, pose)


# TODO mostly duplicate of make_pose_rel_to_parent
def _abs_pose_from_ap_orientation(scene: CScene, project: CProject, pose: Pose, parent_id: str) -> Pose:

    if parent_id in scene.object_ids:
        parent_obj = scene.object(parent_id)
        if not parent_obj.pose:
            raise Arcor2Exception("Parent object does not have pose!")
        parent_pose = parent_obj.pose
    elif parent_id in project.action_points_ids:

        parent_ap = project.bare_action_point(parent_id)

        if parent_ap.parent:
            pose = _abs_pose_from_ap_orientation(scene, project, pose, parent_ap.parent)

        parent_pose = Pose(parent_ap.position, Orientation())

    else:
        raise Arcor2Exception("Unknown parent_id.")

    return make_pose_abs(parent_pose, pose)


def abs_pose_from_ap_orientation(scene: CScene, project: CProject, orientation_id: str) -> Pose:
    """Returns absolute Pose without modifying anything within the project.

    :param orientation_id:
    :return:
    """

    ap, ori = project.bare_ap_and_orientation(orientation_id)
    pose = Pose(ap.position, ori.orientation)

    if not ap.parent:
        return pose

    return _abs_pose_from_ap_orientation(scene, project, pose, ap.parent)



import copy
from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Dict, Iterator, List, Optional, Set, Tuple, ValuesView

from arcor2.data import common as cmn
from arcor2.exceptions import Arcor2Exception

# TODO cached ProjectFunction (actions from functions are totally ignored at the moment)


class CachedSceneException(Arcor2Exception):
    pass


class CachedScene:
    def __init__(self, scene: cmn.Scene):

        self.id: str = scene.id
        self.name: str = scene.name
        self.desc: str = scene.desc
        self.modified: Optional[datetime] = scene.modified
        self.int_modified: Optional[datetime] = scene.int_modified

        # TODO deal with children
        self._objects: Dict[str, cmn.SceneObject] = {}

        for obj in scene.objects:

            if obj.id in self._objects:
                raise CachedSceneException(f"Duplicate object id: {obj.id}.")

            self._objects[obj.id] = obj

    @property
    def bare(self) -> cmn.BareScene:
        return cmn.BareScene(self.id, self.name, self.desc, self.modified, self.int_modified)

    def object_names(self) -> Iterator[str]:

        for obj in self._objects.values():
            yield obj.name

    @property
    def objects(self) -> Iterator[cmn.SceneObject]:

        for obj in self._objects.values():
            yield obj

    @property
    def object_ids(self) -> Set[str]:
        return set(self._objects.keys())

    def object(self, object_id: str) -> cmn.SceneObject:

        try:
            return self._objects[object_id]
        except KeyError:
            raise Arcor2Exception(f"Object ID {object_id} not found.")

    def objects_of_type(self, obj_type: str) -> Iterator[cmn.SceneObject]:

        for obj in self.objects:
            if obj.type == obj_type:
                yield obj

    @property
    def object_types(self) -> Set[str]:
        return {obj.type for obj in self.objects}

    @property
    def scene(self) -> cmn.Scene:

        sc = cmn.Scene.from_bare(self.bare)
        sc.modified = self.modified
        sc.int_modified = self.int_modified
        sc.objects = list(self.objects)
        return sc


class UpdateableCachedScene(CachedScene):
    def __init__(self, scene: cmn.Scene):
        super(UpdateableCachedScene, self).__init__(copy.deepcopy(scene))

    def update_modified(self) -> None:
        self.int_modified = datetime.now(tz=timezone.utc)

    def has_changes(self) -> bool:

        if self.int_modified is None:
            return False

        if self.modified is None:
            return True

        return self.int_modified > self.modified

    def upsert_object(self, obj: cmn.SceneObject) -> None:

        self._objects[obj.id] = obj
        self.update_modified()

    def delete_object(self, obj_id: str) -> None:

        try:
            del self._objects[obj_id]
        except KeyError as e:
            raise Arcor2Exception("Object id not found.") from e

        self.update_modified()


class CachedProjectException(Arcor2Exception):
    pass


@dataclass
class Actions:

    data: Dict[str, cmn.Action] = field(default_factory=dict)
    parent: Dict[str, cmn.BareActionPoint] = field(default_factory=dict)


@dataclass
class Joints:

    data: Dict[str, cmn.ProjectRobotJoints] = field(default_factory=dict)
    parent: Dict[str, cmn.BareActionPoint] = field(default_factory=dict)


@dataclass
class Orientations:

    data: Dict[str, cmn.NamedOrientation] = field(default_factory=dict)
    parent: Dict[str, cmn.BareActionPoint] = field(default_factory=dict)


class CachedProject:
    def __init__(self, project: cmn.Project):

        self.id: str = project.id
        self.name: str = project.name
        self.scene_id: str = project.scene_id
        self.desc: str = project.desc
        self.has_logic: bool = project.has_logic
        self.modified: Optional[datetime] = project.modified
        self._int_modified: Optional[datetime] = project.int_modified

        self._action_points: Dict[str, cmn.BareActionPoint] = {}

        self._actions = Actions()
        self._joints = Joints()
        self._orientations = Orientations()

        self._constants: Dict[str, cmn.ProjectConstant] = {}
        self._logic_items: Dict[str, cmn.LogicItem] = {}
        self._functions: Dict[str, cmn.ProjectFunction] = {}

        self.overrides: Dict[str, List[cmn.Parameter]] = {}

        for ap in project.action_points:

            if ap.id in self._action_points:
                raise CachedProjectException(f"Duplicate AP id: {ap.id}.")

            bare_ap = cmn.BareActionPoint(ap.id, ap.name, ap.position, ap.parent)
            self._action_points[ap.id] = bare_ap

            for ac in ap.actions:

                if ac.id in self._actions.data:
                    raise CachedProjectException(f"Duplicate action id: {ac.id}.")

                self._actions.data[ac.id] = ac
                self._actions.parent[ac.id] = bare_ap

            for joints in ap.robot_joints:

                if joints.id in self._joints.data:
                    raise CachedProjectException(f"Duplicate joints id: {joints.id}.")

                self._joints.data[joints.id] = joints
                self._joints.parent[joints.id] = bare_ap

            for orientation in ap.orientations:

                if orientation.id in self._orientations.data:
                    raise CachedProjectException(f"Duplicate orientation id: {orientation.id}.")

                self._orientations.data[orientation.id] = orientation
                self._orientations.parent[orientation.id] = bare_ap

        for override in project.object_overrides:
            self.overrides[override.id] = override.parameters

        for constant in project.constants:
            self._constants[constant.id] = constant

        for logic_item in project.logic:
            self._logic_items[logic_item.id] = logic_item

        for function in project.functions:
            self._functions[function.id] = function

    @property
    def logic(self) -> ValuesView[cmn.LogicItem]:
        return self._logic_items.values()

    @property
    def constants(self) -> ValuesView[cmn.ProjectConstant]:
        return self._constants.values()

    @property
    def functions(self) -> ValuesView[cmn.ProjectFunction]:
        return self._functions.values()

    @property
    def project(self) -> cmn.Project:

        proj = cmn.Project.from_bare(self.bare)

        for bare_ap in self._action_points.values():

            ap = cmn.ActionPoint.from_bare(bare_ap)
            ap.actions = self.ap_actions(ap.id)
            ap.robot_joints = self.ap_joints(ap.id)
            ap.orientations = self.ap_orientations(ap.id)
            proj.action_points.append(ap)

        proj.object_overrides = [cmn.SceneObjectOverride(k, v) for k, v in self.overrides.items()]
        proj.constants = list(self.constants)
        proj.functions = list(self.functions)
        proj.logic = list(self.logic)
        return proj

    @property
    def bare(self) -> cmn.BareProject:
        return cmn.BareProject(self.id, self.name, self.scene_id, self.desc, self.has_logic)

    @property
    def action_points(self) -> ValuesView[cmn.BareActionPoint]:
        return self._action_points.values()

    @property
    def action_points_with_parent(self) -> List[cmn.BareActionPoint]:
        """Get action points which are relative to something (parent is set).

        :return:
        """

        return [ap for ap in self._action_points.values() if ap.parent]

    @property
    def action_points_names(self) -> Set[str]:
        return {ap.name for ap in self._action_points.values()}

    @property
    def action_points_ids(self) -> Set[str]:
        return set(self._action_points.keys())

    def ap_and_joints(self, joints_id: str) -> Tuple[cmn.BareActionPoint, cmn.ProjectRobotJoints]:

        try:
            return self._joints.parent[joints_id], self._joints.data[joints_id]
        except KeyError:
            raise CachedProjectException("Unknown joints.")

    def joints(self, joints_id: str) -> cmn.ProjectRobotJoints:

        try:
            return self._joints.data[joints_id]
        except KeyError:
            raise CachedProjectException("Unknown joints.")

    def bare_ap_and_orientation(self, orientation_id: str) -> Tuple[cmn.BareActionPoint, cmn.NamedOrientation]:

        try:
            return self._orientations.parent[orientation_id], self._orientations.data[orientation_id]
        except KeyError:
            raise CachedProjectException("Unknown orientation.")

    def ap_orientations(self, ap_id: str) -> List[cmn.NamedOrientation]:

        # TODO come up with something more efficient
        return [
            self._orientations.data[ori_id]
            for ori_id, parent_ap in self._orientations.parent.items()
            if ap_id == parent_ap.id
        ]

    def ap_joints(self, ap_id: str) -> List[cmn.ProjectRobotJoints]:

        # TODO come up with something more efficient
        return [
            self._joints.data[joints_id]
            for joints_id, parent_ap in self._joints.parent.items()
            if ap_id == parent_ap.id
        ]

    def ap_actions(self, ap_id: str) -> List[cmn.Action]:

        # TODO come up with something more efficient
        return [
            self._actions.data[action_id]
            for action_id, parent_ap in self._actions.parent.items()
            if ap_id == parent_ap.id
        ]

    def ap_action_ids(self, ap_id: str) -> Set[str]:
        return {ac.id for ac in self.ap_actions(ap_id)}

    def ap_orientation_names(self, ap_id: str) -> Set[str]:
        return {ori.name for ori in self.ap_orientations(ap_id)}

    def ap_joint_names(self, ap_id: str) -> Set[str]:
        return {joints.name for joints in self.ap_joints(ap_id)}

    def orientation(self, orientation_id: str) -> cmn.NamedOrientation:

        try:
            return self._orientations.data[orientation_id]
        except KeyError:
            raise CachedProjectException("Unknown orientation.")

    def action(self, action_id: str) -> cmn.Action:

        try:
            return self._actions.data[action_id]
        except KeyError:
            raise CachedProjectException("Action not found")

    def action_point_and_action(self, action_id: str) -> Tuple[cmn.BareActionPoint, cmn.Action]:

        try:
            return self._actions.parent[action_id], self._actions.data[action_id]
        except KeyError:
            raise CachedProjectException("Action not found")

    @property
    def actions(self) -> List[cmn.Action]:
        return list(self._actions.data.values())

    def action_ids(self) -> Set[str]:
        return {action.id for action in self.actions}

    def action_user_names(self) -> Set[str]:
        return {action.name for action in self.actions}

    def bare_action_point(self, action_point_id: str) -> cmn.BareActionPoint:

        try:
            return self._action_points[action_point_id]
        except KeyError:
            raise CachedProjectException("Action point not found")

    def action_point(self, action_point_id: str) -> cmn.ActionPoint:

        ap = cmn.ActionPoint.from_bare(self.bare_action_point(action_point_id))
        ap.orientations = self.ap_orientations(action_point_id)
        ap.robot_joints = self.ap_joints(action_point_id)
        return ap

    def logic_item(self, logic_item_id: str) -> cmn.LogicItem:

        try:
            return self._logic_items[logic_item_id]
        except KeyError:
            raise CachedProjectException("LogicItem not found.")

    def constant(self, constant_id: str) -> cmn.ProjectConstant:

        try:
            return self._constants[constant_id]
        except KeyError:
            raise CachedProjectException("Constant not found.")


class UpdateableCachedProject(CachedProject):
    def __init__(self, project: cmn.Project):
        super(UpdateableCachedProject, self).__init__(copy.deepcopy(project))

    def update_modified(self) -> None:
        self._int_modified = datetime.now(tz=timezone.utc)

    @property
    def has_changes(self) -> bool:

        if self._int_modified is None:
            return False

        if self.modified is None:
            return True

        return self._int_modified > self.modified

    def upsert_action(self, ap_id: str, action: cmn.Action) -> None:

        ap = self.bare_action_point(ap_id)

        if action.id in self._actions.data:
            assert self._actions.parent[action.id] == ap
            self._actions.data[action.id] = action
        else:
            self._actions.data[action.id] = action
            self._actions.parent[action.id] = ap
        self.update_modified()

    def remove_action(self, action_id: str) -> cmn.Action:

        try:
            action = self._actions.data.pop(action_id)
            del self._actions.parent[action_id]  # TODO KeyError here might be probably ignored? (it should not happen)
        except KeyError as e:
            raise CachedProjectException("Action not found.") from e
        self.update_modified()
        return action

    def invalidate_joints(self, ap_id: str) -> None:

        for joints in self.ap_joints(ap_id):
            joints.is_valid = False

    def update_ap_position(self, ap_id: str, position: cmn.Position) -> None:

        ap = self.bare_action_point(ap_id)
        ap.position = position
        self.invalidate_joints(ap_id)
        self.update_modified()

    def upsert_orientation(self, ap_id: str, orientation: cmn.NamedOrientation) -> None:

        ap = self.bare_action_point(ap_id)

        if orientation.id in self._orientations.data:
            assert self._orientations.parent[orientation.id] == ap
            self._orientations.data[orientation.id] = orientation
        else:
            self._orientations.data[orientation.id] = orientation
            self._orientations.parent[orientation.id] = ap
        self.update_modified()

    def remove_orientation(self, orientation_id: str) -> cmn.NamedOrientation:

        try:
            ori = self._orientations.data.pop(orientation_id)
            del self._orientations.parent[orientation_id]
        except KeyError as e:
            raise CachedProjectException("Orientation not found.") from e
        self.update_modified()
        return ori

    def upsert_joints(self, ap_id: str, joints: cmn.ProjectRobotJoints) -> None:

        ap = self.bare_action_point(ap_id)

        if joints.id in self._joints.data:
            assert self._joints.parent[joints.id] == ap
            self._joints.data[joints.id] = joints
        else:
            self._joints.data[joints.id] = joints
            self._joints.parent[joints.id] = ap
        self.update_modified()

    def remove_joints(self, joints_id: str) -> cmn.ProjectRobotJoints:

        try:
            joints = self._joints.data.pop(joints_id)
            del self._joints.parent[joints_id]
        except KeyError as e:
            raise CachedProjectException("Joints not found.") from e
        self.update_modified()
        return joints

    def upsert_action_point(
        self, ap_id: str, name: str, position: cmn.Position, parent: Optional[str] = None
    ) -> cmn.BareActionPoint:

        try:
            ap = self.bare_action_point(ap_id)
            ap.name = name
            if position != ap.position:
                self.invalidate_joints(ap_id)
            ap.position = position
            ap.parent = parent
        except CachedProjectException:
            ap = cmn.BareActionPoint(ap_id, name, position, parent)
            self._action_points[ap_id] = ap
        self.update_modified()
        return ap

    def remove_action_point(self, ap_id: str) -> cmn.BareActionPoint:

        ap = self.bare_action_point(ap_id)

        for action in self.ap_actions(ap_id):
            self.remove_action(action.id)

        for joints in self.ap_joints(ap_id):
            self.remove_joints(joints.id)

        for ori in self.ap_orientations(ap_id):
            self.remove_orientation(ori.id)

        del self._action_points[ap_id]
        self.update_modified()
        return ap

    def upsert_logic_item(self, logic_item: cmn.LogicItem) -> None:
        self._logic_items[logic_item.id] = logic_item
        self.update_modified()

    def remove_logic_item(self, logic_item_id: str) -> cmn.LogicItem:

        try:
            logic_item = self._logic_items.pop(logic_item_id)
        except KeyError as e:
            raise CachedProjectException("Logic item not found.") from e
        self.update_modified()
        return logic_item

    def clear_logic(self) -> None:

        self._logic_items.clear()
        self.update_modified()

    def upsert_constant(self, const: cmn.ProjectConstant) -> None:
        self._constants[const.id] = const
        self.update_modified()

    def remove_constant(self, const_id: str) -> cmn.ProjectConstant:

        try:
            const = self._constants.pop(const_id)
        except KeyError as e:
            raise CachedProjectException("Constant not found.") from e
        self.update_modified()
        return const



from dataclasses import dataclass
from typing import List

from dataclasses_jsonschema import JsonSchemaMixin

from arcor2.data.common import Position
from arcor2.exceptions import Arcor2Exception


@dataclass
class MeshFocusAction(JsonSchemaMixin):

    mesh_focus_points: List[Position]
    robot_space_points: List[Position]

    def __post_init__(self) -> None:

        if len(self.mesh_focus_points) != len(self.robot_space_points):
            raise Arcor2Exception("Invalid MeshFocusAction.")



from typing import Any

from dataclasses_jsonschema import ValidationError

from arcor2.cached import CachedProject as CProject
from arcor2.cached import CachedScene as CScene
from arcor2.data.common import Pose
from arcor2.parameter_plugins.base import ParameterPlugin, ParameterPluginException, TypesDict


class RelativePose(Pose):
    pass


class RelativePosePlugin(ParameterPlugin):
    @classmethod
    def type(cls) -> Any:
        return RelativePose

    @classmethod
    def value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> RelativePose:

        param = project.action(action_id).parameter(parameter_id)

        try:
            return RelativePose.from_json(param.value)
        except ValidationError as e:
            raise ParameterPluginException(e)

    @classmethod
    def value_to_json(cls, value: RelativePose) -> str:
        return value.to_json()



# Copyright 2015: Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Taken from https://github.com/openstack/rally/blob/master/rally/common/plugin/info.py

import re
import sys
from typing import Dict

PARAM_OR_RETURNS_REGEX = re.compile(":(?:param|returns)")
RETURNS_REGEX = re.compile(":returns: (?P<doc>.*)", re.S)
PARAM_REGEX = re.compile(r":param (?P<name>[\*\w]+): (?P<doc>.*?)" r"(?:(?=:param)|(?=:return)|(?=:raises)|\Z)", re.S)


def trim(docstring: str) -> str:
    """trim function from PEP-257."""
    if not docstring:
        return ""
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxsize:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)

    # Current code/unittests expects a line return at
    # end of multiline docstrings
    # workaround expected behavior from unittests
    if "\n" in docstring:
        trimmed.append("")

    # Return a single string:
    return "\n".join(trimmed)


def reindent(string: str) -> str:
    return "\n".join(line.strip() for line in string.strip().split("\n"))


def parse_docstring(docstring: str) -> Dict:
    """Parse the docstring into its components.
    :returns: a dictionary of form
              {
                  "short_description": ...,
                  "long_description": ...,
                  "params": [{"name": ..., "doc": ...}, ...],
                  "returns": ...
              }
    """

    short_description = long_description = returns = ""
    params = {}

    if docstring:
        docstring = trim(docstring.lstrip("\n"))

        lines = docstring.split("\n", 1)
        short_description = lines[0]

        if len(lines) > 1:
            long_description = lines[1].strip()

            params_returns_desc = None

            match = PARAM_OR_RETURNS_REGEX.search(long_description)
            if match:
                long_desc_end = match.start()
                params_returns_desc = long_description[long_desc_end:].strip()
                long_description = long_description[:long_desc_end].rstrip()

            if params_returns_desc:
                params = {name: trim(doc) for name, doc in PARAM_REGEX.findall(params_returns_desc)}

                match = RETURNS_REGEX.search(params_returns_desc)
                if match:
                    returns = reindent(match.group("doc"))

    return {  # TODO return rather dataclass?
        "short_description": short_description,
        "long_description": long_description,
        "params": params,
        "returns": returns,
    }



from datetime import datetime

from arcor2.clients import persistent_storage
from arcor2.clients.persistent_storage import ProjectServiceException  # noqa
from arcor2.data.common import IdDescList, Project, ProjectSources, Scene
from arcor2.data.object_type import Mesh, MeshList, Model, Model3dType, ObjectType
from arcor2.helpers import run_in_executor


async def get_mesh(mesh_id: str) -> Mesh:
    return await run_in_executor(persistent_storage.get_mesh, mesh_id)


async def get_meshes() -> MeshList:
    return await run_in_executor(persistent_storage.get_meshes)


async def get_model(model_id: str, model_type: Model3dType) -> Model:
    return await run_in_executor(persistent_storage.get_model, model_id, model_type)


async def put_model(model: Model) -> None:
    await run_in_executor(persistent_storage.put_model, model)


async def delete_model(model_id: str) -> None:
    await run_in_executor(persistent_storage.delete_model, model_id)


async def get_projects() -> IdDescList:
    return await run_in_executor(persistent_storage.get_projects)


async def get_scenes() -> IdDescList:
    return await run_in_executor(persistent_storage.get_scenes)


async def get_project(project_id: str) -> Project:
    return await run_in_executor(persistent_storage.get_project, project_id)


async def get_project_sources(project_id: str) -> ProjectSources:
    return await run_in_executor(persistent_storage.get_project_sources, project_id)


async def get_scene(scene_id: str) -> Scene:
    return await run_in_executor(persistent_storage.get_scene, scene_id)


async def get_object_type(object_type_id: str) -> ObjectType:
    return await run_in_executor(persistent_storage.get_object_type, object_type_id)


async def get_object_type_ids() -> IdDescList:
    return await run_in_executor(persistent_storage.get_object_type_ids)


async def update_project(project: Project) -> datetime:
    return await run_in_executor(persistent_storage.update_project, project)


async def update_scene(scene: Scene) -> datetime:
    return await run_in_executor(persistent_storage.update_scene, scene)


async def update_project_sources(project_sources: ProjectSources) -> None:
    await run_in_executor(persistent_storage.update_project_sources, project_sources)


async def update_object_type(object_type: ObjectType) -> None:
    await run_in_executor(persistent_storage.update_object_type, object_type)


async def delete_object_type(object_type_id: str) -> None:
    await run_in_executor(persistent_storage.delete_object_type, object_type_id)


async def delete_scene(scene_id: str) -> None:
    await run_in_executor(persistent_storage.delete_scene, scene_id)


async def delete_project(project_id: str) -> None:
    await run_in_executor(persistent_storage.delete_project, project_id)


async def save_mesh_file(mesh_id: str, path: str) -> None:
    """Saves mesh file to a given path."""

    await run_in_executor(persistent_storage.save_mesh_file, mesh_id, path)



import json
from typing import Any, List

from arcor2 import transformations as tr
from arcor2.cached import CachedProject as CProject
from arcor2.cached import CachedScene as CScene
from arcor2.data.common import Pose
from arcor2.parameter_plugins.base import ParameterPlugin, TypesDict
from arcor2.parameter_plugins.list import ListParameterPlugin, get_type_name


class PosePlugin(ParameterPlugin):
    @classmethod
    def type(cls) -> Any:
        return Pose

    @classmethod
    def value(cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str) -> Pose:

        action = project.action(action_id)
        param = action.parameter(parameter_id)
        ori_id: str = cls.param_value(param)

        ap, ori = project.bare_ap_and_orientation(ori_id)
        return Pose(ap.position, ori.orientation)

    @classmethod
    def execution_value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> Pose:

        action = project.action(action_id)
        param = action.parameter(parameter_id)
        ori_id: str = cls.param_value(param)
        return tr.abs_pose_from_ap_orientation(scene, project, ori_id)

    @classmethod
    def value_to_json(cls, value: Pose) -> str:
        return value.to_json()

    @classmethod
    def uses_orientation(cls, project: CProject, action_id: str, parameter_id: str, orientation_id: str) -> bool:

        action = project.action(action_id)
        param = action.parameter(parameter_id)
        return orientation_id == cls.param_value(param)


class PoseListPlugin(ListParameterPlugin):
    @classmethod
    def type(cls):
        return List[Pose]

    @classmethod
    def type_name(cls) -> str:
        return get_type_name(PosePlugin)

    @classmethod
    def value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> List[Pose]:

        ret: List[Pose] = []

        ap, action = project.action_point_and_action(action_id)
        parameter = action.parameter(parameter_id)

        for orientation_id in cls.param_value_list(parameter):
            ret.append(Pose(ap.position, project.orientation(orientation_id).orientation))

        return ret

    @classmethod
    def execution_value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> List[Pose]:

        ap, action = project.action_point_and_action(action_id)

        if not ap.parent:
            return cls.value(type_defs, scene, project, action_id, parameter_id)

        parameter = action.parameter(parameter_id)
        ret: List[Pose] = []

        for orientation_id in cls.param_value_list(parameter):
            ret.append(tr.abs_pose_from_ap_orientation(scene, project, orientation_id))

        return ret

    @classmethod
    def value_to_json(cls, value: List[Pose]) -> str:
        return json.dumps([v.to_json() for v in value])

    @classmethod
    def uses_orientation(cls, project: CProject, action_id: str, parameter_id: str, orientation_id: str) -> bool:

        action = project.action(action_id)
        param = action.parameter(parameter_id)

        for ori_id in cls.param_value_list(param):
            if ori_id == orientation_id:
                return True
        return False



import abc
import json
from typing import Any, List, Type

from arcor2.cached import CachedProject as CProject
from arcor2.cached import CachedScene as CScene
from arcor2.data.common import ActionParameter
from arcor2.parameter_plugins.base import ParameterPlugin, ParameterPluginException, TypesDict


def get_type_name(base: Type[ParameterPlugin]) -> str:
    return f"{base.type_name()}_array"


class ListParameterPlugin(ParameterPlugin):
    @classmethod
    @abc.abstractmethod
    def value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> List[Any]:

        val = super(ListParameterPlugin, cls).value(type_defs, scene, project, action_id, parameter_id)

        if not isinstance(val, list):
            raise ParameterPluginException("Not a list!")
        if val and not isinstance(val[0], cls.type()):
            raise ParameterPluginException("List content does not have expected type!")

        return val

    @classmethod
    def param_value_list(cls, param: ActionParameter) -> List[str]:

        lst = json.loads(param.value)

        if not isinstance(lst, list):
            raise ParameterPluginException("Parameter value should be list of references to action points.")

        return lst



import json
from dataclasses import dataclass
from enum import Enum
from typing import Any, Callable, Optional, Set, get_type_hints

from dataclasses_jsonschema import JsonSchemaMixin
from typed_ast import ast3 as ast

from arcor2.cached import CachedProject as CProject
from arcor2.cached import CachedScene as CScene
from arcor2.data.common import IntEnum
from arcor2.data.object_type import ParameterMeta
from arcor2.parameter_plugins.base import ParameterPlugin, ParameterPluginException, TypesDict

# TODO move IntEnum definition here?


@dataclass
class IntegerEnumExtra(JsonSchemaMixin):

    allowed_values: Optional[Set[Any]] = None


class IntegerEnumPlugin(ParameterPlugin):

    EXACT_TYPE = False

    @classmethod
    def type(cls) -> Any:
        return IntEnum

    @classmethod
    def type_name(cls) -> str:
        return "integer_enum"

    @classmethod
    def meta(cls, param_meta: ParameterMeta, action_method: Callable, action_node: ast.FunctionDef) -> None:
        super(IntegerEnumPlugin, cls).meta(param_meta, action_method, action_node)

        ttype = get_type_hints(action_method)[param_meta.name]
        if not issubclass(ttype, cls.type()):
            raise ParameterPluginException(f"Type {ttype.__name__} is not subclass of {cls.type().__name__}.")

        param_meta.extra = IntegerEnumExtra(ttype.set()).to_json()

    @classmethod
    def value(cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str) -> Enum:

        action = project.action(action_id)
        param = action.parameter(parameter_id)
        obj_id, action_type = action.parse_type()
        obj_type = type_defs[scene.object(obj_id).type]

        method = getattr(obj_type, action_type)
        ttype = get_type_hints(method)[param.name]

        if not issubclass(ttype, cls.type()):
            raise ParameterPluginException(f"Type {ttype.__name__} is not subclass of {cls.type().__name__}.")

        try:
            return ttype(json.loads(param.value))
        except ValueError:
            raise ParameterPluginException(f"Parameter {parameter_id} of action {action.name} has invalid value.")



from arcor2.data.common import ActionMetadata
from arcor2.object_types.abstract import GenericWithPose


class Box(GenericWithPose):

    _ABSTRACT = False

    def test(self) -> None:
        """Run test.

        :return:
        """
        pass

    def action_with_unknown_parameter_type(self, param: ActionMetadata) -> bool:
        """Pointless action.

        :param param:
        :return:
        """
        return True

    test.__action__ = ActionMetadata()  # type: ignore
    action_with_unknown_parameter_type.__action__ = ActionMetadata()  # type: ignore



import inspect
import json
from dataclasses import is_dataclass
from typing import Any, Callable, Dict, Iterator, List, Optional, Set, Tuple, Type, get_type_hints

import typing_inspect  # type: ignore
from dataclasses_jsonschema import JsonSchemaMixin, ValidationError
from typed_ast.ast3 import Name

import arcor2
from arcor2.data.common import ActionMetadata, Parameter
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types.abstract import Generic, Settings
from arcor2.source.utils import find_class_def, parse


class ObjectTypeException(Arcor2Exception):
    pass


def check_object_type(type_def: Type[Generic]) -> None:
    """Checks whether the object type source is a valid one.

    :param object_type_source:
    :return:
    """

    if not issubclass(type_def, Generic):
        raise ObjectTypeException("Not a subclass of Generic.")

    # TODO some more (simple) checks here?


def built_in_types() -> Iterator[Tuple[str, Type[Generic]]]:
    """Yields class name and class definition tuple."""

    for cls in inspect.getmembers(arcor2.object_types.abstract, inspect.isclass):
        if not issubclass(cls[1], Generic):
            continue

        yield cls[0], cls[1]


def get_built_in_type(name: str) -> Type[Generic]:

    for bname, cls in built_in_types():
        if name == bname:
            return cls

    raise KeyError


def built_in_types_names() -> Set[str]:

    names = set()

    for type_name, _ in built_in_types():
        names.add(type_name)

    return names


class DataError(Arcor2Exception):
    pass


# TODO settings_to_params


def settings_from_params(
    type_def: Type[Generic], settings: List[Parameter], overrides: Optional[List[Parameter]] = None
) -> Settings:
    """Constructs instance of Settings from two arrays of parameters (scene
    settings and project overrides).

    :param type_def:
    :param settings:
    :param overrides:
    :return:
    """

    if overrides is None:
        overrides = []

    final: Dict[str, Parameter] = {s.name: s for s in settings}
    for over in overrides:
        if over.name not in final:
            raise Arcor2Exception("Invalid override.")

        if over.type != final[over.name].type:
            raise Arcor2Exception("Type mismatch.")

        final[over.name] = over

    settings_def = get_settings_def(type_def)
    settings_data: Dict[str, Any] = {}

    settings_def_type_hints = get_type_hints(settings_def.__init__)

    for s in final.values():

        try:
            setting_def = settings_def_type_hints[s.name]
        except KeyError as e:
            raise Arcor2Exception(f"Unknown property {s.name}.") from e

        try:
            if issubclass(setting_def, JsonSchemaMixin):
                settings_data[s.name] = json.loads(s.value)
            else:
                settings_data[s.name] = setting_def(json.loads(s.value))
        except (ValueError, ValidationError) as e:
            raise Arcor2Exception(f"Parameter {s.name} has invalid value.") from e

    try:
        settings_cls = settings_def.from_dict(settings_data)
    except (ValueError, ValidationError) as e:
        raise Arcor2Exception("Validation of settings failed.") from e

    return settings_cls


def get_settings_def(type_def: Type[Generic]) -> Type[Settings]:
    """Get settings definition from object type definition.

    :param type_def:
    :return:
    """

    sig = inspect.signature(type_def.__init__)
    try:
        param = sig.parameters["settings"]
    except KeyError:
        raise Arcor2Exception("Type has no settings.")

    if typing_inspect.is_optional_type(param.annotation):
        settings_cls = typing_inspect.get_args(param.annotation)[0]
    else:
        settings_cls = param.annotation

    if not issubclass(settings_cls, Settings):
        raise Arcor2Exception("Settings have invalid type.")

    if not is_dataclass(settings_cls):
        raise Arcor2Exception("Settings misses @dataclass decorator.")

    return settings_cls


def base_from_source(source: str, cls_name: str) -> Optional[str]:

    cls_def = find_class_def(cls_name, parse(source))
    if not cls_def.bases:
        return None

    base_name = cls_def.bases[-1]  # allow usage of mixins e.g. class MyType(mixin, Generic)

    assert isinstance(base_name, Name)
    return base_name.id


def iterate_over_actions(
    type_def: Type[Generic],
) -> Iterator[Tuple[str, Callable[[Any,], Any]]]:

    for method_name, method in inspect.getmembers(type_def, inspect.isroutine):

        try:
            if not isinstance(method.__action__, ActionMetadata):
                continue
        except AttributeError:
            continue

        yield method_name, method


__all__ = [
    ObjectTypeException.__name__,
    built_in_types.__name__,
    get_built_in_type.__name__,
    built_in_types_names.__name__,
    DataError.__name__,
]



from typing import Any, Callable, List

from typed_ast import ast3 as ast

from arcor2.cached import CachedProject as CProject
from arcor2.cached import CachedScene as CScene
from arcor2.data.object_type import ParameterMeta
from arcor2.parameter_plugins.base import ParameterPlugin, TypesDict
from arcor2.parameter_plugins.integer import get_min_max
from arcor2.parameter_plugins.list import ListParameterPlugin, get_type_name


class DoublePlugin(ParameterPlugin):
    @classmethod
    def type(cls) -> Any:
        return float

    @classmethod
    def type_name(cls) -> str:
        return "double"

    @classmethod
    def meta(cls, param_meta: ParameterMeta, action_method: Callable, action_node: ast.FunctionDef) -> None:
        super(DoublePlugin, cls).meta(param_meta, action_method, action_node)
        get_min_max(DoublePlugin, param_meta, action_method, action_node)

    @classmethod
    def value(cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str) -> float:
        return cls.type()(super(DoublePlugin, cls).value(type_defs, scene, project, action_id, parameter_id))


class DoubleListPlugin(ListParameterPlugin):
    @classmethod
    def type(cls):
        return List[float]

    @classmethod
    def type_name(cls) -> str:
        return get_type_name(DoublePlugin)

    @classmethod
    def meta(cls, param_meta: ParameterMeta, action_method: Callable, action_node: ast.FunctionDef) -> None:
        super(DoubleListPlugin, cls).meta(param_meta, action_method, action_node)

    @classmethod
    def value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> List[float]:
        return super(DoubleListPlugin, cls).value(type_defs, scene, project, action_id, parameter_id)



import os
from dataclasses import dataclass
from typing import Optional, Set

from dataclasses_jsonschema import JsonSchemaMixin

from arcor2 import rest
from arcor2.data.common import Pose
from arcor2.data.object_type import Model3dType, Models
from arcor2.data.scene import MeshFocusAction
from arcor2.exceptions import Arcor2Exception
from arcor2.exceptions.helpers import handle

URL = os.getenv("ARCOR2_SCENE_SERVICE_URL", "http://0.0.0.0:5013")


class SceneServiceException(Arcor2Exception):
    pass


@dataclass
class MeshParameters(JsonSchemaMixin):

    mesh_scale_x: float = 1.0
    mesh_scale_y: float = 1.0
    mesh_scale_z: float = 1.0
    transform_id: str = "world"


@handle(SceneServiceException, message="Failed to add or update the collision model.")
def upsert_collision(model: Models, pose: Pose, mesh_parameters: Optional[MeshParameters] = None) -> None:
    """Adds arbitrary collision model to the collision scene.

    :param model: Box, Sphere, Cylinder, Mesh
    :param pose: Pose of the collision object.
    :param mesh_parameters: Some additional parameters might be specified for mesh collision model.
    :return:

    Example usage:

    >>> from arcor2.data.object_type import Box
    >>> from arcor2.data.common import Pose, Position, Orientation
    >>> box = Box("UniqueBoxId", 0.1, 0.1, 0.1)
    >>> scene_service.upsert_collision(box, Pose(Position(1, 0, 0), Orientation(0, 0, 0, 1)))
    """

    model_id = model.id
    params = model.to_dict()
    del params["id"]
    params[model.__class__.__name__.lower() + "Id"] = model_id

    if model.type() == Model3dType.MESH and mesh_parameters:
        params.update(mesh_parameters.to_dict())

    rest.call(rest.Method.PUT, f"{URL}/collisions/{model.type().value.lower()}", body=pose, params=params)


@handle(SceneServiceException, message="Failed to delete the collision.")
def delete_collision_id(collision_id: str) -> None:
    rest.call(rest.Method.DELETE, f"{URL}/collisions/{collision_id}")


@handle(SceneServiceException, message="Failed to list collisions.")
def collision_ids() -> Set[str]:
    return set(rest.call(rest.Method.GET, f"{URL}/collisions", list_return_type=str))


@handle(SceneServiceException, message="Failed to focus the object.")
def focus(mfa: MeshFocusAction) -> Pose:
    return rest.call(rest.Method.PUT, f"{URL}/utils/focus", body=mfa, return_type=Pose)


def delete_all_collisions() -> None:

    for cid in collision_ids():
        delete_collision_id(cid)


@handle(SceneServiceException, message="Failed to start the scene.")
def start() -> None:
    """To be called after all objects are created."""

    rest.call(rest.Method.PUT, f"{URL}/system/start")


@handle(SceneServiceException, message="Failed to stop the scene.")
def stop() -> None:
    """To be called when project is closed or when main script ends."""

    rest.call(rest.Method.PUT, f"{URL}/system/stop")


def started() -> bool:
    """Checks whether the scene is running."""

    return rest.call(rest.Method.PUT, f"{URL}/system/running", return_type=bool)


@handle(SceneServiceException, message="Failed to get transforms.")
def transforms() -> Set[str]:
    """Gets available transformations."""

    return set(rest.call(rest.Method.GET, f"{URL}/transforms", list_return_type=str))


@handle(SceneServiceException, message="Failed to add or update the transform.")
def upsert_transform(transform_id: str, parent: str, pose: Pose) -> None:
    """Add or updates transform."""

    rest.call(rest.Method.PUT, f"{URL}/transforms", body=pose, params={"transformId": transform_id, "parent": parent})


@handle(SceneServiceException, message="Failed to get the local pose.")
def local_pose(transform_id: str) -> Pose:
    """Gets relative pose to parent."""

    return rest.call(rest.Method.GET, f"{URL}/transforms/{transform_id}/localPose", return_type=Pose)


@handle(SceneServiceException, message="Failed to get the world pose.")
def world_pose(transform_id: str) -> Pose:
    """Gets absolute pose in world space."""

    return rest.call(rest.Method.GET, f"{URL}/transforms/{transform_id}/worldPose", return_type=Pose)


__all__ = [
    SceneServiceException.__name__,
    upsert_collision.__name__,
    delete_collision_id.__name__,
    collision_ids.__name__,
    focus.__name__,
    delete_all_collisions.__name__,
    start.__name__,
    stop.__name__,
    transforms.__name__,
    upsert_transform.__name__,
    local_pose.__name__,
    world_pose.__name__,
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from arcor2.data.object_type import Box
from arcor2.object_types.upload import upload


def main() -> None:

    upload("arcor2.test_objects/Box", Box("Box", 0.1, 0.1, 0.1))
    upload("arcor2.test_objects/Box2", Box("Box2", 0.2, 0.2, 0.2))
    upload("arcor2.test_objects/Tester", Box("Tester", 0.3, 0.3, 0.3))


if __name__ == "__main__":
    main()



import abc
import json
from typing import Any, Callable, Dict, Type

import humps  # type: ignore
from typed_ast import ast3 as ast

from arcor2.cached import CachedProject, CachedScene
from arcor2.data.common import ActionParameter
from arcor2.data.object_type import ParameterMeta
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types.abstract import Generic

TypesDict = Dict[str, Type[Generic]]


class ParameterPluginException(Arcor2Exception):
    pass


class ParameterPlugin(metaclass=abc.ABCMeta):

    EXACT_TYPE = True

    @classmethod
    @abc.abstractmethod
    def type(cls) -> Any:
        """Returns python type."""
        pass

    @classmethod
    def type_name(cls) -> str:
        """Returns parameter type as string used in JSON."""
        return humps.depascalize(cls.type().__name__)

    @classmethod
    def meta(cls, param_meta: ParameterMeta, action_method: Callable, action_node: ast.FunctionDef) -> None:

        assert param_meta.type == cls.type_name()
        assert param_meta.name

    @classmethod
    def param_value(cls, param: ActionParameter) -> str:

        # TODO support for link, constant (will need access to project)
        assert param.value is not None

        try:
            ret = json.loads(param.value)
        except ValueError:
            raise ParameterPluginException(f"Parameter: {param.name} has invalid value: {param.value}.")

        if not isinstance(ret, str):
            raise ParameterPluginException(
                f"Parameter: {param.name} has invalid value (string expected): {param.value}."
            )

        return ret

    @classmethod
    @abc.abstractmethod
    def value(
        cls, type_defs: TypesDict, scene: CachedScene, project: CachedProject, action_id: str, parameter_id: str
    ) -> Any:

        param = project.action(action_id).parameter(parameter_id)
        try:
            return json.loads(param.value)
        except json.decoder.JSONDecodeError as e:
            raise ParameterPluginException(f"Value of {action_id}/{parameter_id} is not a valid JSON.", e)

    @classmethod
    def execution_value(
        cls, type_defs: TypesDict, scene: CachedScene, project: CachedProject, action_id: str, parameter_id: str
    ) -> Any:

        return cls.value(type_defs, scene, project, action_id, parameter_id)

    @classmethod
    def value_to_json(cls, value: Any) -> str:
        return json.dumps(value)

    @classmethod
    def uses_orientation(cls, project: CachedProject, action_id: str, parameter_id: str, orientation_id: str) -> bool:
        return False

    @classmethod
    def uses_robot_joints(cls, project: CachedProject, action_id: str, parameter_id: str, robot_joints_id: str) -> bool:
        return False



import asyncio
import json
import time
from collections import deque
from typing import Any, Awaitable, Callable, Coroutine, Dict, Optional, Set, Tuple, Type, TypeVar

import websockets
from aiologger.levels import LogLevel  # type: ignore
from dataclasses_jsonschema import ValidationError

from arcor2.data.events import Event
from arcor2.data.rpc.common import RPC
from arcor2.exceptions import Arcor2Exception

RPCT = TypeVar("RPCT", bound=RPC)
ReqT = TypeVar("ReqT", bound=RPC.Request)
RespT = TypeVar("RespT", bound=RPC.Response)

RPC_CB = Callable[[ReqT, websockets.WebSocketServerProtocol], Coroutine[Any, Any, Optional[RespT]]]
RPC_DICT_TYPE = Dict[str, Tuple[Type[RPC], RPC_CB]]

EventT = TypeVar("EventT", bound=Event)
EVENT_DICT_TYPE = Dict[
    str, Tuple[Type[EventT], Callable[[EventT, websockets.WebSocketServerProtocol], Coroutine[Any, Any, None]]]
]


async def send_json_to_client(client: websockets.WebSocketServerProtocol, data: str) -> None:

    try:
        await client.send(data)
    except websockets.exceptions.ConnectionClosed:
        pass


async def server(
    client: Any,
    path: str,
    logger: Any,
    register: Callable[[Any], Awaitable[None]],
    unregister: Callable[[Any], Awaitable[None]],
    rpc_dict: RPC_DICT_TYPE,
    event_dict: Optional[EVENT_DICT_TYPE] = None,
    verbose: bool = False,
) -> None:

    if event_dict is None:
        event_dict = {}

    req_last_ts: Dict[str, deque] = {}
    ignored_reqs: Set[str] = set()

    try:

        await register(client)

        async for message in client:

            try:
                data = json.loads(message)
            except json.decoder.JSONDecodeError as e:
                logger.error(f"Invalid data: '{message}'.")
                logger.debug(e)
                continue

            if not isinstance(data, dict):
                logger.error(f"Invalid data: '{data}'.")
                continue

            if "request" in data:  # ...then it is RPC

                req_type = data["request"]

                try:
                    rpc_cls, rpc_cb = rpc_dict[req_type]
                except KeyError:
                    logger.error(f"Unknown RPC request: {data}.")
                    continue

                assert req_type == rpc_cls.__name__

                try:
                    req = rpc_cls.Request.from_dict(data)
                except ValidationError as e:
                    logger.error(f"Invalid RPC: {data}, error: {e}")
                    continue
                except Arcor2Exception as e:
                    # this might happen if e.g. some dataclass does additional validation of values in its __post_init__
                    try:
                        await client.send(rpc_cls.Response(data["id"], False, messages=[str(e)]).to_json())
                        logger.debug(e, exc_info=True)
                    except KeyError:
                        pass
                    continue

                else:

                    try:
                        resp = await rpc_cb(req, client)
                    except Arcor2Exception as e:
                        logger.debug(e, exc_info=True)
                        resp = rpc_cls.Response(req.id, False, [str(e)])
                    else:
                        if resp is None:  # default response
                            resp = rpc_cls.Response(req.id, True)
                        else:
                            assert isinstance(resp, rpc_cls.Response)
                            resp.id = req.id

                await client.send(resp.to_json())

                if logger.level == LogLevel.DEBUG:

                    # Silencing of repetitive log messages
                    # ...maybe this could be done better and in a more general way using logging.Filter?

                    now = time.monotonic()
                    if req.request not in req_last_ts:
                        req_last_ts[req.request] = deque()

                    while req_last_ts[req.request]:
                        if req_last_ts[req.request][0] < now - 5.0:
                            req_last_ts[req.request].popleft()
                        else:
                            break

                    req_last_ts[req.request].append(now)
                    req_per_sec = len(req_last_ts[req.request]) / 5.0

                    if req_per_sec > 2:
                        if req.request not in ignored_reqs:
                            ignored_reqs.add(req.request)
                            logger.debug(f"Request of type {req.request} will be silenced.")
                    elif req_per_sec < 1:
                        if req.request in ignored_reqs:
                            ignored_reqs.remove(req.request)

                    if req.request not in ignored_reqs:
                        asyncio.ensure_future(logger.debug(f"RPC request: {req}, result: {resp}"))

            elif "event" in data:  # ...event from UI

                try:
                    event_cls, event_cb = event_dict[data["event"]]
                except KeyError as e:
                    logger.error(f"Unknown event type: {e}.")
                    continue

                try:
                    event = event_cls.from_dict(data)
                except ValidationError as e:
                    logger.error(f"Invalid event: {data}, error: {e}")
                    continue

                await event_cb(event, client)

            else:
                logger.error(f"unsupported format of message: {data}")
    except websockets.exceptions.ConnectionClosed:
        pass
    finally:
        await unregister(client)



from typing import Any

from arcor2.data.common import StrEnum
from arcor2.parameter_plugins.integer_enum import IntegerEnumPlugin


class StringEnumPlugin(IntegerEnumPlugin):
    @classmethod
    def type(cls) -> Any:
        return StrEnum

    @classmethod
    def type_name(cls) -> str:
        return "string_enum"



from typing import Dict, List, NamedTuple, Optional, Set, Union

from arcor2.cached import CachedProject
from arcor2.data.common import Action, LogicItem, ProjectFunction
from arcor2.exceptions import Arcor2Exception

LogicContainer = Union[CachedProject, ProjectFunction]
ActionCache = Dict[str, Action]
LogicCache = Dict[str, LogicItem]
ActionIo = Dict[str, List[LogicItem]]

# TODO def validate_logic() -> to be called before saving project / building the package


class ActionCacheTuple(NamedTuple):  # TODO this will be probably obsolete once there will be CachedProjectFunction?

    actions: ActionCache
    logic: LogicCache

    action_inputs: ActionIo
    action_outputs: ActionIo

    first_logic_item: Optional[LogicItem] = None
    last_logic_item: Optional[LogicItem] = None

    @classmethod
    def from_logic_container(cls, container: LogicContainer) -> "ActionCacheTuple":

        actions_cache: ActionCache = {}
        logic_cache: LogicCache = {}
        first_logic_item: Optional[LogicItem] = None
        last_logic_item: Optional[LogicItem] = None

        action_inputs: ActionIo = {}
        action_outputs: ActionIo = {}

        for act in container.actions:
            actions_cache[act.id] = act

        for act in actions_cache.values():
            action_inputs[act.id] = []
            action_outputs[act.id] = []

        valid_endpoints = {LogicItem.START, LogicItem.END} | actions_cache.keys()

        for logic_item in container.logic:

            if logic_item.start not in valid_endpoints:
                raise Arcor2Exception(f"Logic item '{logic_item.id}' has invalid start.")

            if logic_item.end not in valid_endpoints:
                raise Arcor2Exception(f"Logic item '{logic_item.id}' has invalid end.")

            if logic_item.start == logic_item.START:
                if first_logic_item:
                    raise Arcor2Exception("Duplicate start.")
                first_logic_item = logic_item
            elif logic_item.end == logic_item.END:
                if last_logic_item:
                    raise Arcor2Exception("Duplicate end.")
                last_logic_item = logic_item
            else:
                start = logic_item.parse_start()
                action_outputs[start.start_action_id].append(logic_item)
                action_inputs[logic_item.end].append(logic_item)

            logic_cache[logic_item.id] = logic_item

        if first_logic_item and first_logic_item.id not in logic_cache:
            raise Arcor2Exception("Unknown start.")

        if last_logic_item and last_logic_item.id not in logic_cache:
            raise Arcor2Exception("Unknown end.")

        return ActionCacheTuple(
            actions_cache, logic_cache, action_inputs, action_outputs, first_logic_item, last_logic_item
        )


def check_for_loops(parent: LogicContainer, first_action_id: Optional[str] = None) -> None:

    visited_actions: Set[str] = set()
    cache = ActionCacheTuple.from_logic_container(parent)

    def _check_for_loops(action: Action) -> None:

        if action.id in visited_actions:
            raise Arcor2Exception("Loop detected!")

        visited_actions.add(action.id)

        for output in cache.action_outputs[action.id]:
            if output.end == output.END:
                continue

            _check_for_loops(cache.actions[output.end])

    if not first_action_id:
        if not cache.first_logic_item:
            raise Arcor2Exception("Can't check unfinished logic.")

        first_action = cache.actions[cache.first_logic_item.end]
    else:
        first_action = cache.actions[first_action_id]

    _check_for_loops(first_action)



from arcor2.data.common import ActionMetadata
from arcor2.object_types.abstract import Generic


class LogicActions(Generic):
    """Logic-related actions."""

    _ABSTRACT = False

    def equals(self, val1: int, val2: int) -> bool:
        """Tests if two integer values are equal.

        :param val1:
        :param val2:
        :return:
        """
        return val1 == val2

    def less_than(self, val1: int, val2: int) -> bool:
        """Compares two integer values.

        :param val1:
        :param val2:
        :return:
        """
        return val1 < val2

    def greater_than(self, val1: int, val2: int) -> bool:
        """Compares two integer values.

        :param val1:
        :param val2:
        :return:
        """
        return val1 > val2

    equals.__action__ = ActionMetadata(blocking=True)  # type: ignore
    less_than.__action__ = ActionMetadata(blocking=True)  # type: ignore
    greater_than.__action__ = ActionMetadata(blocking=True)  # type: ignore



import time
from typing import List

from arcor2.data.common import ActionMetadata
from arcor2.object_types.abstract import GenericWithPose


class Tester(GenericWithPose):
    """A generic tester."""

    _ABSTRACT = False

    def __init__(self, *args, **kwargs) -> None:
        super(Tester, self).__init__(*args, **kwargs)
        self._cancel: bool = False
        self._param1: str = ""

    def run_test(self, seq: List[float], seq_id: List[str], eqp_res: List[bool]) -> List[bool]:
        """Run test with many parameters.

        :param seq:
        :param seq_id:
        :param eqp_res:
        :return:
        """
        return [bool(seq), bool(seq_id), bool(eqp_res)]

    def long_running_action(self) -> None:
        """This runs for long time.

        :return:
        """
        for _ in range(60):
            time.sleep(1)
            if self._cancel:
                self._cancel = False
                break

    def long_running_action_with_params(self, param1: str) -> None:
        """Runs for long time.

        :param param1:
        :return:
        """
        self._param1 = param1
        for _ in range(60):
            time.sleep(1)
            if self._cancel:
                self._cancel = False
                break

    def simple_cancel(self) -> None:
        self._cancel = True

    def cancel_with_param(self, param1: str) -> None:
        assert param1 == self._param1
        self._cancel = True

    run_test.__action__ = ActionMetadata(blocking=True)  # type: ignore
    long_running_action.__action__ = ActionMetadata(blocking=True)  # type: ignore
    long_running_action_with_params.__action__ = ActionMetadata(blocking=True)  # type: ignore


Tester.CANCEL_MAPPING[Tester.long_running_action.__name__] = Tester.simple_cancel.__name__
Tester.CANCEL_MAPPING[Tester.long_running_action_with_params.__name__] = Tester.cancel_with_param.__name__



from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional

from dataclasses_jsonschema import JsonSchemaMixin

from arcor2.data import common, object_type

"""
------------------------------------------------------------------------------------------------------------------------
Common stuff
------------------------------------------------------------------------------------------------------------------------
"""


@dataclass
class Event(JsonSchemaMixin):
    class Type(common.StrEnum):
        ADD: str = "add"
        UPDATE: str = "update"
        REMOVE: str = "remove"
        UPDATE_BASE: str = "update_base"

    event: str = field(init=False)
    change_type: Optional[Type] = field(init=False)
    parent_id: Optional[str] = field(init=False)

    def __post_init__(self) -> None:
        self.event = self.__class__.__name__
        self.change_type = None
        self.parent_id = None


@dataclass
class Notification(Event):
    @dataclass
    class Data(JsonSchemaMixin):
        class Level(common.StrEnum):
            INFO: str = "Info"
            WARN: str = "Warn"
            ERROR: str = "Error"

        message: str
        level: Level

    data: Optional[Data] = None


"""
------------------------------------------------------------------------------------------------------------------------
Project execution
------------------------------------------------------------------------------------------------------------------------
"""


@dataclass
class ProjectException(Event):
    @dataclass
    class Data(JsonSchemaMixin):
        message: str
        type: str
        handled: bool = False

    data: Data


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class CurrentAction(Event):
    @dataclass
    class Data(JsonSchemaMixin):
        action_id: str = ""
        args: List[common.ActionParameter] = field(default_factory=list)

    data: Data


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class PackageState(Event):
    @dataclass
    class Data(JsonSchemaMixin):
        class StateEnum(Enum):
            RUNNING: str = "running"
            STOPPED: str = "stopped"
            PAUSED: str = "paused"
            UNDEFINED: str = "undefined"

        state: StateEnum = StateEnum.UNDEFINED
        package_id: Optional[str] = None

    RUN_STATES = (Data.StateEnum.PAUSED, Data.StateEnum.RUNNING)

    data: Data


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class PackageInfo(Event):
    @dataclass
    class Data(JsonSchemaMixin):
        package_id: str
        package_name: str
        scene: common.Scene
        project: common.Project
        collision_models: object_type.CollisionModels = field(default_factory=object_type.CollisionModels)

    data: Data


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class ActionState(Event):
    @dataclass
    class Data(JsonSchemaMixin):
        class StateEnum(Enum):
            BEFORE: str = "before"
            AFTER: str = "after"

        object_id: str
        method: str
        where: StateEnum = StateEnum.BEFORE

    data: Data



import sys
import time
import traceback

from arcor2.data.events import ProjectException
from arcor2.exceptions import Arcor2Exception


def format_stacktrace() -> str:
    parts = ["Traceback (most recent call last):\n"]
    parts.extend(traceback.format_stack(limit=25)[:-2])
    parts.extend(traceback.format_exception(*sys.exc_info())[1:])
    return "".join(parts)


def print_exception(e: Exception) -> None:
    """This is intended to be called from the main script. It prints out
    exception in form of JSON (ProjectException).

    The related traceback is saved to a text file for latter diagnosis.
    :param e: Exception to be printed out.
    :return:
    """

    pee = ProjectException(ProjectException.Data(str(e), e.__class__.__name__, isinstance(e, Arcor2Exception)))
    print(pee.to_json())
    sys.stdout.flush()

    with open("traceback-{}.txt".format(time.strftime("%Y%m%d-%H%M%S")), "w") as tb_file:
        tb_file.write(format_stacktrace())



import importlib
import inspect
import pkgutil

from dataclasses_jsonschema import JsonSchemaMixin, ValidationError

from arcor2.exceptions import Arcor2Exception


class DataException(Arcor2Exception):
    pass


def compile_json_schemas() -> None:
    """
    Force compilation of json schema (otherwise it might cause troubles later when executed in parallel)
    :return:
    """

    from arcor2 import data

    for _, module_name, _ in pkgutil.iter_modules(data.__path__):  # type: ignore

        module = importlib.import_module(f"{data.__name__}.{module_name}")

        for _, obj in inspect.getmembers(module, inspect.isclass):

            if not issubclass(obj, JsonSchemaMixin) or obj is JsonSchemaMixin:
                continue

            try:
                obj.from_dict({})
            except ValidationError:
                pass



import functools
from typing import Any, Callable, Optional, Type, TypeVar, cast

from arcor2.exceptions import Arcor2Exception

F = TypeVar("F", bound=Callable[..., Any])


def handle(
    raise_type: Type[Arcor2Exception],
    except_type: Type[Arcor2Exception] = Arcor2Exception,
    message: Optional[str] = None,
) -> Callable[[F], F]:
    def _handle_exceptions(func: F) -> F:
        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> Any:

            try:
                return func(*args, **kwargs)
            except except_type as e:
                if message is not None:
                    raise raise_type(message) from e
                else:
                    raise raise_type(str(e)) from e

        return cast(F, wrapper)

    return _handle_exceptions



import asyncio
import importlib
import keyword
import os
import sys
from threading import Lock
from types import ModuleType
from typing import Any, Callable, Tuple, Type, TypeVar

import humps  # type: ignore
from packaging.version import Version, parse

from arcor2.exceptions import Arcor2Exception


class ImportClsException(Arcor2Exception):
    pass


class TypeDefException(Arcor2Exception):
    pass


def import_cls(module_cls: str) -> Tuple[ModuleType, Type[Any]]:
    """Gets module and class based on string like 'module/Cls'.

    :param module_cls:
    :return:
    """

    try:
        module_name, cls_name = module_cls.split("/")
    except (IndexError, ValueError):
        raise ImportClsException("Invalid format.")

    try:
        module = importlib.import_module(module_name)
    except ModuleNotFoundError:
        raise ImportClsException(f"Module '{module_name}' not found.")

    try:
        cls = getattr(module, cls_name)
    except AttributeError:
        raise ImportClsException(f"Class {cls_name} not found in module '{module_name}'.")

    return module, cls


def is_valid_identifier(value: str) -> bool:
    """
    Identifier (e.g. object id) will be used as variable name in the script - it should be in snake_case,
    not containing any special characters etc.
    :param value:
    :return:
    """

    return value.isidentifier() and not keyword.iskeyword(value) and humps.is_snakecase(value)


def is_valid_type(value: str) -> bool:
    """
    Value will be used as object type name - it should be in CamelCase,
    not containing any special characters etc.
    :param value:
    :return:
    """

    return value.isidentifier() and not keyword.iskeyword(value) and humps.is_pascalcase(value)


S = TypeVar("S")


async def run_in_executor(func: Callable[..., S], *args) -> S:
    return await asyncio.get_event_loop().run_in_executor(None, func, *args)


T = TypeVar("T")


def save_and_import_type_def(source: str, type_name: str, output_type: Type[T], path: str, module_name: str) -> Type[T]:
    """Save source to a file in object_type_path directory.

    :param source:
    :param type_name:
    :param output_type:
    :param object_type_path:
    :return:
    """

    if path not in sys.path:
        sys.path.append(path)

    type_file = humps.depascalize(type_name)
    full_path = os.path.join(path, module_name, type_file)

    with open(f"{full_path}.py", "w") as file:
        file.write(source)

    try:
        module = importlib.import_module(f"{module_name}.{type_file}")
    except ModuleNotFoundError:
        raise ImportClsException(f"Module '{module_name}' not found.")

    try:
        cls = getattr(module, type_name)
    except AttributeError:
        raise ImportClsException(f"Class {type_name} not found in module '{module_name}'.")

    if not issubclass(cls, output_type):
        raise ImportClsException("Not a required type.")

    return cls


def check_compatibility(my_version: str, their_version: str) -> None:

    try:
        mv = parse(my_version)
        tv = parse(their_version)
    except ValueError as e:
        raise Arcor2Exception from e

    if not isinstance(mv, Version):
        raise Arcor2Exception(f"{my_version} is not a proper version.")

    if not isinstance(tv, Version):
        raise Arcor2Exception(f"{their_version} is not a proper version.")

    if mv.major != tv.major:
        raise Arcor2Exception("Different major version.")

    if mv.major == 0:
        if mv.minor != tv.minor:
            raise Arcor2Exception(f"Our version {my_version} is not compatible with {their_version}.")
    else:
        if mv.minor > tv.minor:
            raise Arcor2Exception(f"Our version {my_version} is outdated for {their_version}.")


class NonBlockingLock:
    """This lock can only be used as a context manager."""

    def __init__(self):
        self._lock = Lock()

    def locked(self) -> bool:
        return self._lock.locked()

    def __enter__(self) -> None:
        if not self._lock.acquire(blocking=False):
            raise Arcor2Exception("Can't get the lock.")

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self._lock.release()
        return None


def port_from_url(url: str) -> int:

    return int(url.strip().split(":")[-1])



import pkgutil
import warnings
from typing import Dict, NamedTuple, Set


class DynamicParamTuple(NamedTuple):

    method_name: str
    parent_parameters: Set[str]


# key: name of parameter, value: name of method to call (to get set of strings), set of parent parameters
DynamicParamDict = Dict[str, DynamicParamTuple]

# key: action, value: cancel method
CancelDict = Dict[str, str]

warnings.filterwarnings(action="ignore", category=UserWarning, module="quaternion")

warnings.filterwarnings(
    action="ignore", category=UserWarning, module="dataclasses_jsonschema", message="Unable to create schema for 'Any'"
)


def package_version(package: str, file: str = "VERSION") -> str:

    res = pkgutil.get_data(__name__, file)
    if not res:
        return "unknown"
    return res.decode().strip()


def version() -> str:
    return package_version(__name__)



import os
from datetime import datetime

from arcor2 import rest
from arcor2.data.common import IdDescList, Project, ProjectSources, Scene
from arcor2.data.object_type import MODEL_MAPPING, Mesh, MeshList, Model, Model3dType, ObjectType
from arcor2.exceptions import Arcor2Exception
from arcor2.exceptions.helpers import handle

URL = os.getenv("ARCOR2_PERSISTENT_STORAGE_URL", "http://0.0.0.0:11000")

# TODO thread to poll changes? how to "detect" changes?


class ProjectServiceException(Arcor2Exception):
    pass


@handle(ProjectServiceException, message="Failed to get the mesh.")
def get_mesh(mesh_id: str) -> Mesh:
    return rest.call(rest.Method.GET, f"{URL}/models/{mesh_id}/mesh", return_type=Mesh)


@handle(ProjectServiceException, message="Failed to get list of meshes.")
def get_meshes() -> MeshList:
    return rest.call(rest.Method.GET, f"{URL}/models/meshes", list_return_type=Mesh)


@handle(ProjectServiceException, message="Failed to get the model type.")
def get_model(model_id: str, model_type: Model3dType) -> Model:
    return rest.call(
        rest.Method.GET, f"{URL}/models/{model_id}/{model_type.value.lower()}", return_type=MODEL_MAPPING[model_type]
    )


@handle(ProjectServiceException, message="Failed to add or update the model.")
def put_model(model: Model) -> None:
    rest.call(rest.Method.PUT, f"{URL}/models/{model.__class__.__name__.lower()}", body=model)


@handle(ProjectServiceException, message="Failed to delete the model.")
def delete_model(model_id: str) -> None:
    rest.call(rest.Method.DELETE, f"{URL}/models/{model_id}")


@handle(ProjectServiceException, message="Failed to list projects.")
def get_projects() -> IdDescList:
    return rest.call(rest.Method.GET, f"{URL}/projects", return_type=IdDescList)


@handle(ProjectServiceException, message="Failed to list scenes.")
def get_scenes() -> IdDescList:
    return rest.call(rest.Method.GET, f"{URL}/scenes", return_type=IdDescList)


@handle(ProjectServiceException, message="Failed to get the project.")
def get_project(project_id: str) -> Project:
    return rest.call(rest.Method.GET, f"{URL}/project/{project_id}", return_type=Project)


@handle(ProjectServiceException, message="Failed to get the project sources.")
def get_project_sources(project_id: str) -> ProjectSources:
    return rest.call(rest.Method.GET, f"{URL}/project/{project_id}/sources", return_type=ProjectSources)


@handle(ProjectServiceException, message="Failed to get the scene.")
def get_scene(scene_id: str) -> Scene:
    return rest.call(rest.Method.GET, f"{URL}/scene/{scene_id}", return_type=Scene)


@handle(ProjectServiceException, message="Failed to get the object type.")
def get_object_type(object_type_id: str) -> ObjectType:
    return rest.call(rest.Method.GET, f"{URL}/object_types/{object_type_id}", return_type=ObjectType)


@handle(ProjectServiceException, message="Failed to list object types.")
def get_object_type_ids() -> IdDescList:
    return rest.call(rest.Method.GET, f"{URL}/object_types", return_type=IdDescList)


@handle(ProjectServiceException, message="Failed to add or update the project.")
def update_project(project: Project) -> datetime:

    assert project.id
    return datetime.fromisoformat(rest.call(rest.Method.PUT, f"{URL}/project", return_type=str, body=project))


@handle(ProjectServiceException, message="Failed to add or update the scene.")
def update_scene(scene: Scene) -> datetime:

    assert scene.id
    return datetime.fromisoformat(rest.call(rest.Method.PUT, f"{URL}/scene", return_type=str, body=scene))


@handle(ProjectServiceException, message="Failed to add or update the project sources.")
def update_project_sources(project_sources: ProjectSources) -> None:

    assert project_sources.id
    rest.call(rest.Method.POST, f"{URL}/project/sources", body=project_sources)


@handle(ProjectServiceException, message="Failed to add or update the object type.")
def update_object_type(object_type: ObjectType) -> None:

    assert object_type.id
    rest.call(rest.Method.PUT, f"{URL}/object_type", body=object_type)


@handle(ProjectServiceException, message="Failed to delete the object type.")
def delete_object_type(object_type_id: str) -> None:
    rest.call(rest.Method.DELETE, f"{URL}/object_type/{object_type_id}")


@handle(ProjectServiceException, message="Failed to delete the scene.")
def delete_scene(scene_id: str) -> None:
    rest.call(rest.Method.DELETE, f"{URL}/scene/{scene_id}")


@handle(ProjectServiceException, message="Failed to delete the project.")
def delete_project(project_id: str) -> None:
    rest.call(rest.Method.DELETE, f"{URL}/project/{project_id}")


@handle(ProjectServiceException, message="Failed to get the mesh.")
def save_mesh_file(mesh_id: str, path: str) -> None:
    """Saves mesh file to a given path."""

    rest.download(f"{URL}/models/{mesh_id}/mesh/file", path)



import json
import logging
import os
from enum import Enum
from functools import partial
from io import BytesIO
from typing import Any, Dict, List, NamedTuple, Optional, Sequence, Type, TypeVar, Union, overload

import humps  # type: ignore
import requests
from dataclasses_jsonschema import JsonSchemaMixin, ValidationError
from PIL import Image, UnidentifiedImageError  # type: ignore

from arcor2.exceptions import Arcor2Exception
from arcor2.logging import get_logger

# typing-related definitions
DataClass = TypeVar("DataClass", bound=JsonSchemaMixin)
Primitive = TypeVar("Primitive", str, int, float, bool)
ReturnValue = Union[None, Primitive, DataClass, BytesIO]
ReturnType = Union[None, Type[Primitive], Type[DataClass], Type[BytesIO]]

OptBody = Optional[Union[JsonSchemaMixin, Sequence[JsonSchemaMixin], Sequence[Primitive]]]
OptParams = Optional[Dict[str, Primitive]]
OptFiles = Optional[Dict[str, bytes]]


class RestException(Arcor2Exception):
    """Exception raised by functions in the module."""

    pass


class Method(Enum):
    """Enumeration of supported HTTP methods."""

    GET = partial(requests.get)
    POST = partial(requests.post)
    PUT = partial(requests.put)
    DELETE = partial(requests.delete)
    PATCH = partial(requests.patch)


class Timeout(NamedTuple):
    """The connect timeout is the number of seconds Requests will wait for your
    client to establish a connection to a remote machine (corresponding to the
    connect()) call on the socket. It’s a good practice to set connect timeouts
    to slightly larger than a multiple of 3, which is the default TCP packet
    retransmission window.

    Once your client has connected to the server and sent the HTTP request,
    the read timeout is the number of seconds the client will wait for the server
    to send a response. (Specifically, it’s the number of seconds that the client
    will wait between bytes sent from the server. In 99.9% of cases, this is
    the time before the server sends the first byte).

    Source: https://requests.readthedocs.io/en/master/user/advanced/#timeouts
    """

    connect: float = 3.05
    read: float = 20.0


OptTimeout = Optional[Timeout]


# module-level variables
debug: bool = bool(os.getenv("ARCOR2_REST_DEBUG", False))
headers = {"accept": "application/json", "content-type": "application/json"}
session = requests.session()
logger = get_logger(__name__, logging.DEBUG if debug else logging.INFO)


def dataclass_from_json(resp_json: Dict[str, Any], return_type: Type[DataClass]) -> DataClass:

    try:
        return return_type.from_dict(resp_json)
    except ValidationError as e:
        logger.debug(f'{return_type.__name__}: validation error "{e}" while parsing "{resp_json}".')
        raise RestException("Invalid data.", str(e)) from e


def primitive_from_json(resp_json: Primitive, return_type: Type[Primitive]) -> Primitive:

    try:
        return return_type(resp_json)
    except ValueError as e:
        logger.debug(f'{return_type.__name__}: error  "{e}" while parsing "{resp_json}".')
        raise RestException(e) from e


# overload for no return
@overload
def call(
    method: Method,
    url: str,
    *,
    body: OptBody = None,
    params: OptParams = None,
    files: OptFiles = None,
    timeout: OptTimeout = None,
) -> None:
    ...


# single value-returning overloads
@overload
def call(
    method: Method,
    url: str,
    *,
    return_type: Type[Primitive],
    body: OptBody = None,
    params: OptParams = None,
    files: OptFiles = None,
    timeout: OptTimeout = None,
) -> Primitive:
    ...


@overload
def call(
    method: Method,
    url: str,
    *,
    return_type: Type[DataClass],
    body: OptBody = None,
    params: OptParams = None,
    files: OptFiles = None,
    timeout: OptTimeout = None,
) -> DataClass:
    ...


@overload
def call(
    method: Method,
    url: str,
    *,
    return_type: Type[BytesIO],
    body: OptBody = None,
    params: OptParams = None,
    files: OptFiles = None,
    timeout: OptTimeout = None,
) -> BytesIO:
    ...


# list-returning overloads
@overload
def call(
    method: Method,
    url: str,
    *,
    list_return_type: Type[Primitive],
    body: OptBody = None,
    params: OptParams = None,
    files: OptFiles = None,
    timeout: OptTimeout = None,
) -> List[Primitive]:
    ...


@overload
def call(
    method: Method,
    url: str,
    *,
    list_return_type: Type[DataClass],
    body: OptBody = None,
    params: OptParams = None,
    files: OptFiles = None,
    timeout: OptTimeout = None,
) -> List[DataClass]:
    ...


@overload
def call(
    method: Method,
    url: str,
    *,
    list_return_type: Type[BytesIO],
    body: OptBody = None,
    params: OptParams = None,
    files: OptFiles = None,
    timeout: OptTimeout = None,
) -> List[BytesIO]:
    ...


def call(
    method: Method,
    url: str,
    *,
    return_type: ReturnType = None,
    list_return_type: ReturnType = None,
    body: OptBody = None,
    params: OptParams = None,
    files: OptFiles = None,
    timeout: OptTimeout = None,
) -> ReturnValue:
    """Universal function for calling REST APIs.

    :param method: HTTP method.
    :param url: Resource address.
    :param return_type: If set, function will try to return one value of a given type.
    :param list_return_type: If set, function will try to return list of a given type.
    :param body: Data to be send in the request body.
    :param params: Path parameters.
    :param files: Instead of body, it is possible to send files.
    :param timeout: Specific timeout for a call.
    :return: Return value/type is given by return_type/list_return_type. If both are None, nothing will be returned.
    """
    logger.debug(f"{method} {url}, body: {body}, params: {params}, files: {files is not None}, timeout: {timeout}")

    if body and files:
        raise RestException("Can't send data and files at the same time.")

    if return_type and list_return_type:
        raise RestException("Only one argument from 'return_type' and 'list_return_type' can be used.")

    if return_type is None:
        return_type = list_return_type

    # prepare data into dict
    if isinstance(body, JsonSchemaMixin):
        d = humps.camelize(body.to_dict())
    elif isinstance(body, list):
        d = []
        for dd in body:
            if isinstance(dd, JsonSchemaMixin):
                d.append(humps.camelize(dd.to_dict()))
            else:
                d.append(dd)
    elif body is not None:
        raise RestException("Unsupported type of data.")
    else:
        d = {}

    if params:
        params = humps.camelize(params)
    else:
        params = {}

    if timeout is None:
        timeout = Timeout()

    try:
        if files:
            resp = method.value(url, files=files, timeout=timeout, params=params)
        else:
            resp = method.value(url, data=json.dumps(d), timeout=timeout, headers=headers, params=params)
    except requests.exceptions.RequestException as e:
        logger.debug("Request failed.", exc_info=True)
        raise RestException("Catastrophic system error.", str(e)) from e

    _handle_response(resp)

    if return_type is None:
        return None

    if issubclass(return_type, BytesIO):

        if list_return_type:
            raise NotImplementedError

        return BytesIO(resp.content)

    logger.debug(f"Response text: {resp.text}")

    try:
        resp_json = resp.json()
    except ValueError as e:
        logger.debug(f"Got invalid JSON in the response: {resp.text}")
        raise RestException("Invalid JSON.") from e

    logger.debug(f"Response json: {resp_json}")
    if isinstance(resp_json, (dict, list)):
        resp_json = humps.decamelize(resp_json)
    logger.debug(f"Decamelized json: {resp_json}")

    if list_return_type and not isinstance(resp_json, list):
        logger.debug(f"Expected list of type {return_type}, but got {resp_json}.")
        raise RestException("Response is not a list.")

    if issubclass(return_type, JsonSchemaMixin):

        if list_return_type:
            return [dataclass_from_json(item, return_type) for item in resp_json]

        else:
            assert not isinstance(resp_json, list)

            # TODO temporary workaround for bug in humps (https://github.com/nficano/humps/issues/127)
            from arcor2.data.object_type import Box

            if return_type is Box:
                resp_json["size_x"] = resp_json["sizex"]
                resp_json["size_y"] = resp_json["sizey"]
                resp_json["size_z"] = resp_json["sizez"]

            return dataclass_from_json(resp_json, return_type)

    else:  # probably a primitive

        if list_return_type:
            return [primitive_from_json(item, return_type) for item in resp_json]
        else:
            assert not isinstance(resp_json, list)
            return primitive_from_json(resp_json, return_type)


def _handle_response(resp: requests.Response) -> None:
    """Raises exception if there is something wrong with the response.

    :param resp:
    :return:
    """

    try:
        resp.raise_for_status()
    except requests.exceptions.RequestException as e:

        try:
            resp_body = json.loads(resp.content)
        except json.JSONDecodeError:
            raise RestException(resp.content.decode("utf-8")) from e

        try:
            raise RestException(resp_body["message"]) from e
        except (KeyError, TypeError):  # TypeError is for case when resp_body is just string
            raise RestException(str(resp_body)) from e


def get_image(url: str) -> Image.Image:
    """Shortcut for getting an image."""

    # TODO check content type?
    try:
        return Image.open(call(Method.GET, url, return_type=BytesIO))
    except (UnidentifiedImageError, TypeError) as e:
        raise RestException("Invalid image.") from e


def download(url: str, path: str, params: OptParams = None) -> None:
    """Shortcut for saving a file to disk."""

    with call(Method.GET, url, return_type=BytesIO, params=params) as buff:
        with open(path, "wb") as file:
            file.write(buff.getvalue())



from typing import Dict, Set, Type

# TODO just temporary solution - allow loading plugins based on env. var or from project service?
from arcor2.parameter_plugins.base import ParameterPlugin
from arcor2.parameter_plugins.boolean import BooleanListPlugin, BooleanPlugin
from arcor2.parameter_plugins.double import DoubleListPlugin, DoublePlugin
from arcor2.parameter_plugins.image import ImagePlugin
from arcor2.parameter_plugins.integer import IntegerListPlugin, IntegerPlugin
from arcor2.parameter_plugins.integer_enum import IntegerEnumPlugin
from arcor2.parameter_plugins.joints import JointsPlugin
from arcor2.parameter_plugins.pose import PoseListPlugin, PosePlugin
from arcor2.parameter_plugins.relative_pose import RelativePosePlugin
from arcor2.parameter_plugins.string import StringListPlugin, StringPlugin
from arcor2.parameter_plugins.string_enum import StringEnumPlugin

Plugins = Dict[Type, Type[ParameterPlugin]]

PLUGINS: Set[Type[ParameterPlugin]] = {
    BooleanPlugin,
    BooleanListPlugin,
    DoublePlugin,
    DoubleListPlugin,
    ImagePlugin,
    IntegerPlugin,
    IntegerListPlugin,
    IntegerEnumPlugin,
    JointsPlugin,
    PosePlugin,
    PoseListPlugin,
    RelativePosePlugin,
    StringPlugin,
    StringListPlugin,
    StringEnumPlugin,
}

PARAM_PLUGINS: Dict[str, Type[ParameterPlugin]] = {}
TYPE_TO_PLUGIN: Plugins = {}

for plug in PLUGINS:
    if plug.type_name() in PARAM_PLUGINS:
        print(f"Plugin for type {plug.type_name()} already registered.")
        continue
    # TODO figure out why mypy complains
    # Only concrete class can be given where "Type[ParameterPlugin]" is expected
    PARAM_PLUGINS[plug.type_name()] = plug  # type: ignore
    TYPE_TO_PLUGIN[plug.type()] = plug  # type: ignore



from dataclasses import dataclass, field
from typing import List, NamedTuple, Optional

from dataclasses_jsonschema import JsonSchemaMixin


class Arcor2Mixin(JsonSchemaMixin):
    @classmethod
    def get_qualname(cls) -> str:
        return cls.__qualname__.split(".")[0]


class RPC:
    """Base class for all RPCs."""

    @dataclass
    class Request(Arcor2Mixin):  # TODO from_dict -> check that request key is == to qualname

        id: int
        request: str = field(init=False)

        def __post_init__(self) -> None:
            self.request = self.get_qualname()

    @dataclass
    class Response(Arcor2Mixin):

        id: int = 0
        response: str = field(init=False)
        result: bool = True
        messages: Optional[List[str]] = None
        # TODO define data here somehow? And check that if result==True there are some data

        def __post_init__(self) -> None:
            self.response = self.get_qualname()


# ----------------------------------------------------------------------------------------------------------------------


@dataclass
class TypeArgs(JsonSchemaMixin):
    type: str


@dataclass
class IdArgs(JsonSchemaMixin):
    id: str


@dataclass
class RobotArg(JsonSchemaMixin):
    class RobotArgTuple(NamedTuple):

        robot_id: str
        end_effector: str

    robot_id: str = field(metadata=dict(description="Object id of the robot or robot_id within the robot service."))
    end_effector: str

    def as_tuple(self) -> RobotArgTuple:
        return self.RobotArgTuple(self.robot_id, self.end_effector)


# ----------------------------------------------------------------------------------------------------------------------


class Version(RPC):
    @dataclass
    class Request(RPC.Request):
        pass

    @dataclass
    class Response(RPC.Response):
        @dataclass
        class Data(JsonSchemaMixin):
            version: str = ""

        data: Optional[Data] = None



import base64
import io
import json
from typing import Any

import PIL.Image  # type: ignore
from PIL.Image import Image  # type: ignore

from arcor2.cached import CachedProject as CProject
from arcor2.cached import CachedScene as CScene
from arcor2.parameter_plugins.base import ParameterPlugin, TypesDict


class ImagePlugin(ParameterPlugin):
    @classmethod
    def type(cls) -> Any:
        return Image

    @classmethod
    def type_name(cls) -> str:
        return "image"

    @classmethod
    def value(cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str) -> Image:
        json_str = super(ImagePlugin, cls).value(type_defs, scene, project, action_id, parameter_id)
        b64_bytes = json_str.encode()
        image_data = base64.b64decode(b64_bytes)
        return PIL.Image.open(io.BytesIO(image_data))

    @classmethod
    def value_to_json(cls, value: Image) -> str:

        with io.BytesIO() as output:
            value.save(output, "jpeg")
            b64_bytes = base64.b64encode(output.getvalue())
            b64_str = b64_bytes.decode()
            return json.dumps(b64_str)



import importlib
import inspect
from typing import List, Optional, Type, Union

import autopep8  # type: ignore
import horast
import typed_astunparse
from typed_ast.ast3 import (
    AST,
    Assert,
    Attribute,
    Call,
    ClassDef,
    Expr,
    FunctionDef,
    ImportFrom,
    Load,
    Module,
    Name,
    NodeTransformer,
    NodeVisitor,
    Raise,
    Store,
    alias,
    fix_missing_locations,
)

from arcor2.source import SourceException


def parse(source: str) -> AST:

    try:
        return horast.parse(source)
    except SyntaxError as e:
        raise SourceException("Failed to parse the code.") from e


def parse_def(type_def: Type) -> AST:
    try:
        return parse(inspect.getsource(type_def))
    except OSError as e:
        raise SourceException("Failed to get the source code.") from e


def find_asserts(tree: FunctionDef) -> List[Assert]:
    class FindAsserts(NodeVisitor):
        def __init__(self) -> None:
            self.asserts: List[Assert] = []

        def visit_Assert(self, node: Assert) -> None:
            self.asserts.append(node)

    ff = FindAsserts()
    ff.visit(tree)

    return ff.asserts


def find_function(name: str, tree: Union[Module, AST]) -> FunctionDef:
    class FindFunction(NodeVisitor):
        def __init__(self) -> None:
            self.function_node: Optional[FunctionDef] = None

        def visit_FunctionDef(self, node: FunctionDef) -> None:
            if node.name == name:
                self.function_node = node
                return

            if not self.function_node:
                self.generic_visit(node)

    ff = FindFunction()
    ff.visit(tree)

    if ff.function_node is None:
        raise SourceException(f"Function {name} not found.")

    return ff.function_node


def find_class_def(name: str, tree: Union[Module, AST]) -> ClassDef:
    class FindClassDef(NodeVisitor):
        def __init__(self) -> None:
            self.cls_def_node: Optional[ClassDef] = None

        def visit_ClassDef(self, node: ClassDef) -> None:
            if node.name == name:
                self.cls_def_node = node
                return

            if not self.cls_def_node:
                self.generic_visit(node)

    ff = FindClassDef()
    ff.visit(tree)

    if ff.cls_def_node is None:
        raise SourceException(f"Class definition {name} not found.")

    return ff.cls_def_node


def add_import(node: Module, module: str, cls: str, try_to_import: bool = True) -> None:
    """Adds "from ... import ..." to the beginning of the script.

    Parameters
    ----------
    node
    module
    cls

    Returns
    -------
    """

    class AddImportTransformer(NodeTransformer):
        def __init__(self, module: str, cls: str) -> None:
            self.done = False
            self.module = module
            self.cls = cls

        def visit_ImportFrom(self, node: ImportFrom) -> ImportFrom:
            if node.module == self.module:

                for aliass in node.names:
                    if aliass.name == self.cls:
                        self.done = True
                        break
                else:
                    node.names.append(alias(name=self.cls, asname=None))
                    self.done = True

            return node

    if try_to_import:

        try:
            imported_mod = importlib.import_module(module)
        except ModuleNotFoundError as e:
            raise SourceException(e)

        try:
            getattr(imported_mod, cls)
        except AttributeError as e:
            raise SourceException(e)

    tr = AddImportTransformer(module, cls)
    node = tr.visit(node)

    if not tr.done:
        node.body.insert(0, ImportFrom(module=module, names=[alias(name=cls, asname=None)], level=0))


def append_method_call(body: List, instance: str, method: str, args: List, kwargs: List) -> None:
    body.append(Expr(value=Call(func=get_name_attr(instance, method), args=args, keywords=kwargs)))


def get_name(name: str) -> Name:
    return Name(id=name, ctx=Load())


def get_name_attr(name: str, attr: str, ctx: Union[Type[Load], Type[Store]] = Load) -> Attribute:
    return Attribute(value=get_name(name), attr=attr, ctx=ctx())


def tree_to_str(tree: AST) -> str:
    # TODO why this fails?
    # validator.visit(tree)

    fix_missing_locations(tree)
    generated_code: str = horast.unparse(tree)
    generated_code = autopep8.fix_code(generated_code, options={"aggressive": 1})

    return generated_code


def dump(tree: Module) -> str:
    return typed_astunparse.dump(tree)


def find_raises(tree: FunctionDef) -> List[Raise]:
    class FindRaises(NodeVisitor):
        def __init__(self) -> None:
            self.raises: List[Raise] = []

        def visit_Raise(self, node: Raise) -> None:
            self.raises.append(node)

    ff = FindRaises()
    ff.visit(tree)

    return ff.raises



import select
import sys
from functools import wraps
from typing import Any, Callable, Type, TypeVar, Union, cast

from arcor2.data.events import ActionState, Event, PackageState
from arcor2.object_types.abstract import Generic
from arcor2.object_types.utils import iterate_over_actions

HANDLE_ACTIONS = True


def patch_object_actions(type_def: Type[Generic]) -> None:
    """Dynamically adds @action decorator to the methods with assigned
    ActionMetadata.

    :param type_def:
    :return:
    """

    for method_name, method in iterate_over_actions(type_def):
        setattr(type_def, method_name, action(method))


try:
    # Windows solution
    import msvcrt
    import time

    def read_stdin(timeout: float = 0.0) -> Union[str, None]:

        time_to_end = time.monotonic() + timeout

        while True:

            x = msvcrt.kbhit()  # type: ignore
            if x:
                return msvcrt.getch().decode()  # type: ignore

            if not timeout or time.monotonic() > time_to_end:
                return None

            time.sleep(timeout / 100.0)


except ImportError:

    # Linux solution
    def read_stdin(timeout: float = 0.0) -> Union[str, None]:

        if select.select([sys.stdin], [], [], timeout)[0]:
            return sys.stdin.readline().strip()
        return None


def handle_action(inst: Generic, f: Callable[..., Any], where: ActionState.Data.StateEnum) -> None:

    print_event(ActionState(ActionState.Data(inst.id, f.__name__, where)))

    ctrl_cmd = read_stdin()

    if ctrl_cmd == "p":
        print_event(PackageState(PackageState.Data(PackageState.Data.StateEnum.PAUSED)))
        while True:
            ctrl_cmd = read_stdin(0.1)
            if ctrl_cmd == "r":
                print_event(PackageState(PackageState.Data(PackageState.Data.StateEnum.RUNNING)))
                break


def print_event(event: Event) -> None:
    """Used from main script to print event as JSON."""

    print(event.to_json())
    sys.stdout.flush()


F = TypeVar("F", bound=Callable[..., Any])


def action(f: F) -> F:
    @wraps(f)
    def wrapper(*args: Union[Generic, Any], **kwargs: Any) -> Any:

        # automagical overload for dictionary (allow to get rid of ** in script).
        if len(args) == 2 and isinstance(args[1], dict) and not kwargs:
            kwargs = args[1]
            args = (args[0],)

        if not action.inside_composite and HANDLE_ACTIONS:  # type: ignore
            handle_action(args[0], f, ActionState.Data.StateEnum.BEFORE)

        if wrapper.__action__.composite:  # type: ignore # TODO and not step_into
            action.inside_composite = f  # type: ignore

        res = f(*args, **kwargs)

        if action.inside_composite == f:  # type: ignore
            action.inside_composite = None  # type: ignore

        if not action.inside_composite and HANDLE_ACTIONS:  # type: ignore
            handle_action(args[0], f, ActionState.Data.StateEnum.AFTER)

        return res

    return cast(F, wrapper)


action.inside_composite = None  # type: ignore



from typing import Any, List

from arcor2.cached import CachedProject as CProject
from arcor2.cached import CachedScene as CScene
from arcor2.parameter_plugins.base import ParameterPlugin, TypesDict
from arcor2.parameter_plugins.list import ListParameterPlugin, get_type_name


class StringPlugin(ParameterPlugin):
    @classmethod
    def type(cls) -> Any:
        return str

    @classmethod
    def type_name(cls) -> str:
        return "string"

    @classmethod
    def value(cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str) -> str:
        return cls.type()(super(StringPlugin, cls).value(type_defs, scene, project, action_id, parameter_id))


class StringListPlugin(ListParameterPlugin):
    @classmethod
    def type(cls):
        return List[str]

    @classmethod
    def type_name(cls) -> str:
        return get_type_name(StringPlugin)

    @classmethod
    def value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> List[str]:
        return super(StringListPlugin, cls).value(type_defs, scene, project, action_id, parameter_id)




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: src/python/arcor2:arcor2_dist

from setuptools import setup

setup(**{
    'author': 'Robo@FIT',
    'author_email': 'imaterna@fit.vut.cz',
    'classifiers': [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 3.8',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering',
    ],
    'description': 'ARCOR2 - Augmented Reality Collaborative Robot.',
    'license': 'LGPL',
    'long_description': """# arcor2
# Changelog

The format is based on [Keep a Changelog](https://keepachangelog.com/en/1.0.0/),

## [0.9.0] - 2020-10-16

### Changed
- Scene client updated for Scene 0.3.0.
- Exceptions refactored (BREAKING).
  - Arcor2Exception no longer has `message` property
  - There is a new package arcor2/exceptions.
  - Clients using `handle` decorator now have specific error messages defined.
  - Previously generated execution packages will become broken as `print_exception` function was moved 
- Loggers are now created using functions from `logging` module.
- `rest` module rewritten (BREAKING).
  - Now there is only one method `call` which takes http method as an argument.
  - There is `ARCOR2_REST_DEBUG` environment variable. When set, debugging logs are turned on.

### Fixed
- JSON containing only boolean value was not handled properly by `arcor2.rest` module.


## [0.8.0] - 2020-09-24
### Changed
- Reorganisation of the repository - switched to monorepo based on [Pants](https://www.pantsbuild.org/docs/welcome-to-pants). The code was divided into more packages (that can be separatelly relased) within one repository.
- Tests now run on GitHub instead of CircleCi.
- Unification of objects and services
  - There is ```Generic``` base class for objects without pose, ```GenericWithPose``` for objects with pose and ```Robot``` class that should be base class for every robot.
- Integration of scene service (0.2.0).
- @action decorator is now added automatically in the run-time.
- ```Orientation``` dataclass now performs quaternion normalization in ```__post_init__```.
- ```Robot``` base class now has ```_move_lock``` mutex to ensure that only one move-action is called at the time.

## [0.8.0rc2] - 2020-09-16

## [0.8.0rc1] - 2020-09-15

## [0.8.0b8] - 2020-08-21
### Fixed
- Some robot-related issues fixed

## [0.8.0b7] - 2020-08-12
### Changed
- Scene service client: 'upsert_collision' now has optional 'mesh_parameters': parameter.

## [0.8.0b6] - 2020-08-03
### Changed
- New logic representation
- Unification of objects and services
- Integration of scene service

## [0.7.1] - 2020-07-15
### Fixed
- Fix of broken python package arcor2 0.7.0

## [0.7.0] - 2020-07-15
### Changed

- ARServer: new RPC 'TemporaryPackage'
- ARServer: RPC ObjectTypesChangedEvent renamed to ChangedObjectTypesEvent, now contains changed ObjectTypes meta instead of just type names
- ARServer: ShowMainScreenEvent.
- Package name added to PackageInfoEvent
- ARServer now compares its API_VERSION with API_VERSION of the Execution.
- ARServer: ShowMainScreenEvent will now not contain 'highlight' when sent to a newly connected UI.
- AP can now have another AP as parent
- rest: OptionalData now may contain list of primitive type.
- Execution: PackageStateEvent now contains package_id
- Execution: added 'executed' to PackageMeta

## [0.6.0] - 2020-06-19
### Changed

- Build/Execution proxy: allow port change using env. var.
- ARServer: RenameScene RPC now checks if scene name is unique and 'dry_run' works.
- ARServer: ListScenes/ListProjects now contain 'modified'.
- ARServer: DeleteObjectType RPC added.
- @action decorator is now compatible with Windows.
- Service class now has 'cleanup' method which is called when scene is closed or when script ends. Call of 'cleanup' can be disabled by 'export ARCOR_CLEANUP_SERVICES=False' - this is particularly useful when running the script manually again and again.
- Cleanup method for ObjectTypes.

## [0.5.1] - 2020-06-04
### Fixed
- ignoring check of return parameters
- allowing list of strings as request body

## [0.5.0] - 2020-06-01
### Changed
- ARServer container need to setup new env variable using docker-compose -> ARCOR2_DATA_PATH=/root/data
- ListProjects RPC now gets projects in parallel.
- dry_run parameter for selected RPCs
- EEF pose/robot joints streaming
- OpenScene, OpenProject, SceneClosed, ProjectClosed events.
- Execution proxy: use persistent websocket connection.
- SceneCollisionsEvent merged into PackageInfoEvent
- ARServer: RPC to cancel action execution.
- Execution package now contains package.json with its metadata. Execution service now supports renaming of packages.

## [0.4.3] - 2020-05-22
### Changed
- added support for CORS

## [0.4.2] - 2020-04-27
### Fixed
- Fix of functions to transform relative poses to absolute and vice versa

## [0.4.1] - 2020-04-22
### Added
- New RPCs for getting robot joints and effector pose
- New RPC to get IDs of EE and suctions
- Added pivot enum for UpdateObjectPoseUsingRobot

### Fixed
- Fix of remove action RPC
- Another fixes


## [0.4.0] - 2020-04-17
### Changed
- Complete redesign of RPC for ARClients (AREditor atm)
- Documentation of execution and build API
- Support for project service 0.2.0
- New and updated events
- Enhanced error messages
- Create (global) session to enable reuse of connections.

## [0.3.0] - 2020-03-24
### Changed
- Renamed RobotJoints to ProjectRobotJoints and ModelTypeEnum to Model3dType
- Added new services for Time and Logic related actons
- Added boolean parameter plugin
- Description, returns and origins fields marked as optional
- New event - ActionResult
- Separated script enabling discovery through UDP broadcast
- Support for list params
- Services and Action objects are now marked as disabled when some problem occured and error message is passed to GUI (previously such services/objects were ignored)
- Services with no configuration are disabled

## [0.2.1] - 2020-02-28
### Fixed
- Added compatibility with Project service v 0.1.1
- Param values and defaults are strings now
- min, max stored in extra as JSON string

## [0.2.0] - 2020-02-20
### Changed
- ExecuteAction RPC.
- Uuid for action object/point/action.
- Execution proxy PUT method
- ActionPoint class in execution package
- Removed loop in main script, when hasLogic == false
- Parameter values not send in currentAction event
- ProjectState RESUMED removed
- Execution: print out script output if not JSON.
- Joint: rotation -> value


## [0.1.7] - 2019-12-24
### Fixed
- Build: disable caching

## [0.1.5] - 2019-12-22
### Fixed
- Parameter plugins

## [0.1.4] - 2019-12-18
### Fixed
- Parameter of type relative_pose now accepts json string as value

## [0.1.3] - 2019-12-18
### Fixed
- N/A

## [0.1.2] - 2019-12-17
### Fixed
- bump docker version

## [0.1.1] - 2019-12-17
### Fixed
- bump docker version

## [0.1.1] - 2019-12-12
### Fixed
- N/A

## [0.1.0] - 2019-12-12
### Changed
- Separation of services.""",
    'long_description_content_type': 'text/markdown',
    'name': 'arcor2',
    'python_requires': '==3.8.*',
    'version': '0.9.0',
    'package_dir': {
        '': 'src',
    },
    'packages': (
        'arcor2',
        'arcor2.clients',
        'arcor2.data',
        'arcor2.data.rpc',
        'arcor2.exceptions',
        'arcor2.object_types',
        'arcor2.parameter_plugins',
        'arcor2.scripts',
        'arcor2.source',
        'arcor2.test_objects',
    ),
    'namespace_packages': (
    ),
    'package_data': {
        'arcor2': (
            'VERSION',
            'py.typed',
        ),
    },
    'install_requires': (
        'Flask==1.1.2',
        'Pillow==7.2.0',
        'PyYAML==5.3.1',
        'aiologger==0.6.0',
        'autopep8==1.5.4',
        'dataclasses-jsonschema-rereleased[fast-validation]==2.12.2',
        'dataclasses==0.6',
        'horast-rereleased==0.4.1',
        'numpy-quaternion==2020.9.5.14.42.2',
        'pyhumps==1.6.1',
        'requests==2.24.0',
        'static-typing==0.2.7',
        'typed-ast==1.4.1',
        'typed-astunparse==2.1.4',
        'typing-inspect==0.6.0',
        'websockets==8.1',
    ),
    'entry_points': {
        'console_scripts': [
            'arcor2_upload_builtin_objects=arcor2.scripts.upload_builtin_objects',
        ],
    },
})



import os
import sys
from datetime import datetime, timezone

from dataclasses_jsonschema import ValidationError

from arcor2.data.execution import PackageMeta

"""
Functions used by the main script itself (Resources class) or Execution service.
"""

PROJECT_PATH_NAME = "ARCOR2_PROJECT_PATH"

try:
    PROJECT_PATH = os.environ[PROJECT_PATH_NAME]
except KeyError:
    sys.exit(f"'{PROJECT_PATH_NAME}' env. variable not set.")


def get_package_meta_path(package_id: str) -> str:

    return os.path.join(PROJECT_PATH, package_id, "package.json")


def read_package_meta(package_id: str) -> PackageMeta:

    try:
        with open(get_package_meta_path(package_id)) as pkg_file:
            return PackageMeta.from_json(pkg_file.read())
    except (IOError, ValidationError):
        return PackageMeta("N/A", datetime.fromtimestamp(0, tz=timezone.utc))


def write_package_meta(package_id: str, meta: PackageMeta) -> None:

    with open(get_package_meta_path(package_id), "w") as pkg_file:
        pkg_file.write(meta.to_json())



from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from dataclasses_jsonschema import JsonSchemaMixin


@dataclass
class PackageMeta(JsonSchemaMixin):

    name: str
    built: datetime
    executed: Optional[datetime] = None



import importlib
import json
import os
from typing import Any, Dict, Optional, Type, TypeVar

import humps  # type: ignore
from dataclasses_jsonschema import JsonSchemaMixin, JsonSchemaValidationError

import arcor2.object_types
from arcor2 import package
from arcor2 import transformations as tr
from arcor2.action import patch_object_actions, print_event
from arcor2.cached import CachedProject, CachedScene
from arcor2.clients import scene_service
from arcor2.data.common import Project, Scene
from arcor2.data.events import CurrentAction, PackageInfo
from arcor2.data.object_type import Box, Cylinder, Mesh, Models, ObjectModel, Sphere
from arcor2.exceptions import Arcor2Exception
from arcor2.exceptions.runtime import print_exception
from arcor2.object_types.abstract import Generic, GenericWithPose, Robot
from arcor2.object_types.utils import built_in_types_names, settings_from_params
from arcor2.parameter_plugins import PARAM_PLUGINS
from arcor2.parameter_plugins.base import TypesDict


class ResourcesException(Arcor2Exception):
    pass


class IntResources:

    CUSTOM_OBJECT_TYPES_MODULE = "object_types"

    def __init__(self, scene: Scene, project: Project, models: Dict[str, Optional[Models]]) -> None:

        self.project = CachedProject(project)
        self.scene = CachedScene(scene)

        if self.project.scene_id != self.scene.id:
            raise ResourcesException("Project/scene not consistent!")

        self.objects: Dict[str, Generic] = {}

        self.type_defs: TypesDict = {}

        built_in = built_in_types_names()

        # workaround for possible 'An item with the same key has already been added.' problem
        scene_service.stop()

        scene_service.delete_all_collisions()

        package_id = os.path.basename(os.getcwd())
        package_meta = package.read_package_meta(package_id)
        package_info_event = PackageInfo(PackageInfo.Data(package_id, package_meta.name, scene, project))

        for scene_obj in self.scene.objects:

            if scene_obj.type in built_in:
                module = importlib.import_module(arcor2.object_types.__name__ + "." + humps.depascalize(scene_obj.type))
            else:
                module = importlib.import_module(
                    ResourcesBase.CUSTOM_OBJECT_TYPES_MODULE + "." + humps.depascalize(scene_obj.type)
                )

            cls = getattr(module, scene_obj.type)
            patch_object_actions(cls)
            self.type_defs[cls.__name__] = cls

            assert scene_obj.id not in self.objects, "Duplicate object id {}!".format(scene_obj.id)

            settings = settings_from_params(cls, scene_obj.parameters, self.project.overrides.get(scene_obj.id, None))

            if issubclass(cls, Robot):
                self.objects[scene_obj.id] = cls(scene_obj.id, scene_obj.name, scene_obj.pose, settings)
            elif issubclass(cls, GenericWithPose):
                self.objects[scene_obj.id] = cls(
                    scene_obj.id, scene_obj.name, scene_obj.pose, models[scene_obj.type], settings
                )
            elif issubclass(cls, Generic):
                self.objects[scene_obj.id] = cls(scene_obj.id, scene_obj.name, settings)
            else:
                raise Arcor2Exception("Unknown base class.")

        for model in models.values():

            if not model:
                continue

            if isinstance(model, Box):
                package_info_event.data.collision_models.boxes.append(model)
            elif isinstance(model, Sphere):
                package_info_event.data.collision_models.spheres.append(model)
            elif isinstance(model, Cylinder):
                package_info_event.data.collision_models.cylinders.append(model)
            elif isinstance(model, Mesh):
                package_info_event.data.collision_models.meshes.append(model)

        scene_service.start()

        print_event(package_info_event)

    def make_all_poses_absolute(self) -> None:
        """This is only needed when the main script is written fully manually
        (actions are not defined in AR Editor).

        :return:
        """

        # make all poses absolute
        for aps in self.project.action_points_with_parent:
            # Action point pose is relative to its parent object/AP pose in scene but is absolute during runtime.
            tr.make_relative_ap_global(self.scene, self.project, aps)

    def __enter__(self) -> "IntResources":
        return self

    def __exit__(self, ex_type, ex_value, traceback) -> bool:

        if ex_type:  # TODO ignore when script is stopped correctly (e.g. KeyboardInterrupt, ??)
            print_exception(ex_type(ex_value))

        scene_service.stop()
        scene_service.delete_all_collisions()

        for obj in self.objects.values():
            obj.cleanup()

        return True

    def print_info(self, action_id: str, args: Dict[str, Any]) -> None:
        """Helper method used to print out info about the action going to be
        executed."""
        # TODO to be used for parameters that are result of previous action(s)
        # ...there is no need to send parameters that are already in the project
        print_event(CurrentAction(CurrentAction.Data(action_id)))

    def parameters(self, action_id: str) -> Dict[str, Any]:

        try:
            act = self.project.action(action_id)
        except Arcor2Exception:
            raise ResourcesException("Action_id {} not found in project {}.".format(action_id, self.project.id))

        ret: Dict[str, Any] = {}

        for param in act.parameters:
            ret[param.name] = PARAM_PLUGINS[param.type].execution_value(
                self.type_defs, self.scene, self.project, action_id, param.name
            )

        return ret


T = TypeVar("T", bound=JsonSchemaMixin)


def lower(s: str) -> str:

    return s.lower()


class ResourcesBase(IntResources):
    def read_project_data(self, file_name: str, cls: Type[T]) -> T:

        try:

            with open(os.path.join("data", file_name + ".json")) as scene_file:
                return cls.from_dict(humps.decamelize(json.loads(scene_file.read())))  # type: ignore

        except JsonSchemaValidationError as e:
            raise ResourcesException(f"Invalid project/scene: {e}")

    def __init__(self, project_id: str) -> None:

        scene = self.read_project_data(Scene.__name__.lower(), Scene)
        project = self.read_project_data(Project.__name__.lower(), Project)

        if project_id != project.id:
            raise ResourcesException("Resources were generated for different project!")

        models: Dict[str, Optional[Models]] = {}

        for obj in scene.objects:

            if obj.type in models:
                continue

            try:
                models[obj.type] = self.read_project_data("models/" + humps.depascalize(obj.type), ObjectModel).model()
            except IOError:
                models[obj.type] = None

        super(ResourcesBase, self).__init__(scene, project, models)



from dataclasses import dataclass
from typing import Any, Callable, List, Tuple, Type

from dataclasses_jsonschema import JsonSchemaMixin
from typed_ast import ast3 as ast

from arcor2.cached import CachedProject as CProject
from arcor2.cached import CachedScene as CScene
from arcor2.data.object_type import ParameterMeta
from arcor2.parameter_plugins.base import ParameterPlugin, ParameterPluginException, TypesDict
from arcor2.parameter_plugins.list import ListParameterPlugin, get_type_name
from arcor2.source.utils import find_asserts


@dataclass
class IntegerParameterExtra(JsonSchemaMixin):

    minimum: Any
    maximum: Any


class AssertNotFound(ParameterPluginException):
    pass


def get_assert_minimum_maximum(asserts: List[ast.Assert], param_name: str) -> Tuple[Any, Any]:

    # Assert(test=Compare(left=Num(n=0), ops=[LtE(), LtE()], comparators=[Name(id='speed', ctx=Load()), Num(n=100)]))

    for ass in asserts:
        if not isinstance(ass.test, ast.Compare):
            continue

        if len(ass.test.comparators) != 2:
            continue

        if len(ass.test.ops) != 2:
            continue

        err = False
        for op in ass.test.ops:
            if not isinstance(op, ast.LtE):
                err = True
                break
        if err:
            continue

        try:
            if ass.test.comparators[0].name == param_name:  # type: ignore
                return ass.test.left.n, ass.test.comparators[1].n  # type: ignore
        except AttributeError:
            continue

    raise AssertNotFound()


def get_min_max(
    cls: Type[ParameterPlugin], param_meta: ParameterMeta, action_method: Callable, action_node: ast.FunctionDef
) -> None:

    try:
        minimum, maximum = get_assert_minimum_maximum(find_asserts(action_node), param_meta.name)
    except AssertNotFound:
        return

    for var in minimum, maximum:
        if not isinstance(var, cls.type()):
            raise ParameterPluginException("Parameter bounds has incorrect type.")

    param_meta.extra = IntegerParameterExtra(minimum, maximum).to_json()


class IntegerPlugin(ParameterPlugin):
    @classmethod
    def type(cls) -> Any:
        return int

    @classmethod
    def type_name(cls) -> str:
        return "integer"

    @classmethod
    def meta(cls, param_meta: ParameterMeta, action_method: Callable, action_node: ast.FunctionDef) -> None:

        super(IntegerPlugin, cls).meta(param_meta, action_method, action_node)
        get_min_max(IntegerPlugin, param_meta, action_method, action_node)

    @classmethod
    def value(cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str) -> int:
        return cls.type()(super(IntegerPlugin, cls).value(type_defs, scene, project, action_id, parameter_id))


class IntegerListPlugin(ListParameterPlugin):
    @classmethod
    def type(cls):
        return List[int]

    @classmethod
    def type_name(cls) -> str:
        return get_type_name(IntegerPlugin)

    @classmethod
    def meta(cls, param_meta: ParameterMeta, action_method: Callable, action_node: ast.FunctionDef) -> None:
        super(IntegerListPlugin, cls).meta(param_meta, action_method, action_node)

    @classmethod
    def value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> List[int]:
        return super(IntegerListPlugin, cls).value(type_defs, scene, project, action_id, parameter_id)



#!/usr/bin/env python3

from arcor2.object_types.logic_actions import LogicActions
from arcor2.object_types.time_actions import TimeActions
from arcor2.object_types.upload import upload_def


def main() -> None:

    upload_def(LogicActions)
    upload_def(TimeActions)


if __name__ == "__main__":
    main()



from typing import Any, List

from arcor2.cached import CachedProject as CProject
from arcor2.cached import CachedScene as CScene
from arcor2.parameter_plugins.base import ParameterPlugin, TypesDict
from arcor2.parameter_plugins.list import ListParameterPlugin, get_type_name


class BooleanPlugin(ParameterPlugin):
    @classmethod
    def type(cls) -> Any:
        return bool

    @classmethod
    def type_name(cls) -> str:
        return "boolean"

    @classmethod
    def value(cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str) -> bool:
        return cls.type()(super(BooleanPlugin, cls).value(type_defs, scene, project, action_id, parameter_id))


class BooleanListPlugin(ListParameterPlugin):
    @classmethod
    def type(cls):
        return List[bool]

    @classmethod
    def type_name(cls) -> str:
        return get_type_name(BooleanPlugin)

    @classmethod
    def value(
        cls, type_defs: TypesDict, scene: CScene, project: CProject, action_id: str, parameter_id: str
    ) -> List[bool]:
        return super(BooleanListPlugin, cls).value(type_defs, scene, project, action_id, parameter_id)



import static_typing as st
import typed_ast.ast3

from arcor2.exceptions import Arcor2Exception

SCRIPT_HEADER = "#!/usr/bin/env python3\n" "# -*- coding: utf-8 -*-\n\n"
validator = st.ast_manipulation.AstValidator[typed_ast.ast3](mode="strict")


class SourceException(Arcor2Exception):
    pass



import asyncio
from typing import Set

from arcor2.clients import scene_service
from arcor2.clients.scene_service import SceneServiceException
from arcor2.data.common import Pose
from arcor2.data.object_type import Models
from arcor2.data.scene import MeshFocusAction
from arcor2.helpers import run_in_executor


async def upsert_collision(model: Models, pose: Pose) -> None:
    await run_in_executor(scene_service.upsert_collision, model, pose)


async def delete_collision_id(collision_id: str) -> None:
    await run_in_executor(scene_service.delete_collision_id, collision_id)


async def collision_ids() -> Set[str]:
    return await run_in_executor(scene_service.collision_ids)


async def focus(mfa: MeshFocusAction) -> Pose:
    return await run_in_executor(scene_service.focus, mfa)


async def start() -> None:
    await run_in_executor(scene_service.start)


async def stop() -> None:
    await run_in_executor(scene_service.stop)


async def started() -> bool:
    return await run_in_executor(scene_service.started)


async def delete_all_collisions() -> None:
    await asyncio.gather(*[delete_collision_id(coll_id) for coll_id in await collision_ids()])


__all__ = [
    upsert_collision.__name__,
    delete_collision_id.__name__,
    collision_ids.__name__,
    focus.__name__,
    delete_all_collisions.__name__,
    SceneServiceException.__name__,
    start.__name__,
    stop.__name__,
    started.__name__,
]



from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, Optional, Set, Type, Union

from dataclasses_jsonschema import JsonSchemaMixin

from arcor2.data import DataException
from arcor2.data.common import Pose


class Model3dType(Enum):

    NONE: str = "None"
    BOX: str = "Box"
    CYLINDER: str = "Cylinder"
    SPHERE: str = "Sphere"
    MESH: str = "Mesh"


@dataclass
class MetaModel3d(JsonSchemaMixin):

    id: str
    type: Model3dType

    def __post_init__(self) -> None:

        if isinstance(self.type, str):
            self.type = Model3dType[self.type.upper()]


@dataclass
class Model(JsonSchemaMixin):

    id: str

    @classmethod
    def type(cls) -> Model3dType:
        return Model3dType[cls.__name__.upper()]

    def metamodel(self) -> MetaModel3d:
        return MetaModel3d(self.id, self.type())


@dataclass
class Box(Model):

    size_x: float
    size_y: float
    size_z: float

    def __post_init__(self) -> None:

        dims = [self.size_x, self.size_y, self.size_z]

        if [val for val in dims if val < 0]:
            raise DataException("Dimensions has to be positive.")

        if len([val for val in dims if val > 0]) < 2:
            raise DataException("Only one dimension of a box can be zero.")


@dataclass
class Cylinder(Model):

    radius: float
    height: float

    def __post_init__(self) -> None:

        if [val for val in [self.radius, self.height] if val <= 0]:
            raise DataException("Dimensions has to be positive.")


@dataclass
class Sphere(Model):
    radius: float

    def __post_init__(self) -> None:

        if self.radius <= 0:
            raise DataException("Radius has to be positive.")


@dataclass
class Mesh(Model):

    uri: str
    focus_points: Optional[List[Pose]] = None


MeshList = List[Mesh]

Models = Union[Box, Sphere, Cylinder, Mesh]

MODEL_MAPPING: Dict[Model3dType, Union[Type[Box], Type[Sphere], Type[Cylinder], Type[Mesh]]] = {
    Box.type(): Box,
    Sphere.type(): Sphere,
    Cylinder.type(): Cylinder,
    Mesh.type(): Mesh,
}


@dataclass
class ObjectType(JsonSchemaMixin):
    """Object type, as it is stored in DB."""

    id: str
    source: str
    desc: Optional[str] = ""
    model: Optional[MetaModel3d] = None

    def __post_init__(self) -> None:  # TODO workaround for bug (?) in Storage

        if self.model and self.model.type == Model3dType.NONE:
            self.model = None


@dataclass
class ObjectModel(JsonSchemaMixin):

    type: Model3dType
    box: Optional[Box] = None
    cylinder: Optional[Cylinder] = None
    sphere: Optional[Sphere] = None
    mesh: Optional[Mesh] = None

    def model(self) -> Models:

        assert self.type != Model3dType.NONE
        return getattr(self, str(self.type.value).lower())

    def __post_init__(self) -> None:

        models_list = [self.box, self.cylinder, self.sphere, self.mesh]

        if models_list.count(None) != len(models_list) - 1:
            raise DataException("No model specified!")


@dataclass
class CollisionModels(JsonSchemaMixin):

    boxes: List[Box] = field(default_factory=list)
    spheres: List[Sphere] = field(default_factory=list)
    cylinders: List[Cylinder] = field(default_factory=list)
    meshes: List[Mesh] = field(default_factory=list)


@dataclass
class ParameterMeta(JsonSchemaMixin):

    name: str
    type: str
    dynamic_value: bool = False  # client should ask for allowed values using RPC
    dynamic_value_parents: Optional[Set[str]] = None
    description: Optional[str] = None
    default_value: Optional[str] = None
    extra: Optional[str] = None
    children: List["ParameterMeta"] = field(default_factory=list)
